/*
								+-----------------------------------+
								|                                   |
								|  ***  LRD classifier engine  ***  |
								|                                   |
								|   Copyright   -tHE SWINe- 2007   |
								|                                   |
								|          Classifier.cpp           |
								|                                   |
								+-----------------------------------+
*/

#include <vector>
#include "../../berLame_src/CallStack.h"
#ifdef WIN32
#include <windows.h> // gl.h needs some typedefs
#endif
#define GL_GLEXT_LEGACY
#include <GL/gl.h>
#include <math.h>
#include <algorithm>
#include <string>
#include "../../berLame_src/OpenGL20.h"
#include "../../berLame_src/OpenGLState.h"
#include "../../berLame_src/Shader2.h"
#include "../../berLame_src/ShaderSys.h"
#include "../../berLame_src/Texture.h"
#include "../../berLame_src/RenderBuffer2.h"
#include "../../berLame_src/Timer.h"
#include "Load_XML.h"
#include "ConvPlan.h"
#include "RasCalib.h"
#include "Classifier.h"
#include "ScopeGuard.h"

#if defined(_MSC_VER) && !defined(__MWERKS__) && !defined(for)
#define for if(0) {} else for
#endif

extern bool Save_Gray_TGA(const char *p_s_file, int n_width, int n_height,
	const unsigned char *p_buffer);
// debug

/*
 *								=== CGPU_AdaBoost::TClassifierData ===
 */

/*
 *	CGPU_AdaBoost::TClassifierData::TClassifierData(CGLState *p_state,
 *		int n_image_width, int n_image_height, const TWeakClassifier *_p_classifier,
 *		CConvolutionManager &r_conv_plan)
 *		- default constructor
 *		- p_state is OpenGL state guard (needed for rasterizer calibration)
 *		- n_image_width and n_image_height are input image width and height
 *		- _p_classifier contains weak classifier parameters as loaded from xml
 *		- r_conv_plan is convolution plan (contains convolution image neccessary
 *		  to evaluate for this classifier)
 *		- if any error occured, convolution image is set to 0 (p_Convolution() returns 0)
 */
CGPU_AdaBoost::TClassifierData::TClassifierData(CGLState *p_state, int n_image_width,
	int n_image_height, const TWeakClassifier *_p_classifier, CConvolutionManager &r_conv_plan)
	:p_convolution(0), p_classifier(_p_classifier)
{
	for(int i = 0; i < r_conv_plan.n_Convolution_Num(); ++ i) {
		if(r_conv_plan.r_Convolution_Image(i).n_block_width == _p_classifier->n_block_width &&
		   r_conv_plan.r_Convolution_Image(i).n_block_height == _p_classifier->n_block_height) {
			p_convolution = r_conv_plan.r_Convolution_Image(i).p_texture;
			break;
		}
	}
	if(!p_convolution)
		return;
	// finds proper convolution image

	float f_off_s = (float)_p_classifier->n_position_x / n_image_width;
	float f_off_t = (float)_p_classifier->n_position_y / n_image_height;
	CRasterizerCalibrator::TTexCoordRect t_orig_texcoords(f_off_s, f_off_t,
		1 - 1.0f / n_image_width + f_off_s, 1 - 1.0f / n_image_height + f_off_t);
	// texture coordinates

	t_texcoords = t_orig_texcoords;
	if(!CRasterizerCalibrator::CalibrateRasterizer(p_state,
	   n_image_width, n_image_height, t_orig_texcoords, t_texcoords, 1e-6f) &&
	   !CRasterizerCalibrator::CalibrateRasterizer(p_state,
	   n_image_width, n_image_height, t_orig_texcoords, t_texcoords, 1e-4f) &&
	   !CRasterizerCalibrator::CalibrateRasterizer(p_state,
	   n_image_width, n_image_height, t_orig_texcoords, t_texcoords, 1e-3f)) { // try lower precision
		p_convolution = 0; // mark error
		return;
	}
	// calibrate 'em so shader gets what it needs

	int n_slice_num = _p_classifier->n_block_width * _p_classifier->n_block_height;
	v_block_to_slice_x = (float)n_image_width / n_slice_num;
	v_block_to_slice_x_modulo =
		(float)_p_classifier->n_block_width / n_slice_num;
	v_block_to_slice_y = (float)n_image_height / n_slice_num *
		_p_classifier->n_block_width;
	v_block_to_slice_z_offset = .1f / (n_slice_num * (n_image_width / 256));
	// calculate slice selection params for the shader

	v_pixel_x = 4.0f / (n_image_width / _p_classifier->n_block_width);
	v_pixel_y = 1.0f / (n_image_height / _p_classifier->n_block_height);
	// calculate size of 3D texture's pixels

	for(int n_pass = 0; n_pass < 4; ++ n_pass) {
		unsigned char p_mask[4] = {0};
		for(unsigned int i = 0; i < 32; ++ i) {
			int x = (i + _p_classifier->n_position_x) / _p_classifier->n_block_width;
			if(x % 4 == n_pass)
				p_mask[i / 8] |= (unsigned)0x80 >> (i % 8);
		}
		// determine horizontal mask profile

		for(int i = 0; i < 32 * 4; ++ i)
			p_stipple_mask[n_pass][i] = p_mask[i % 4];
		// copy verticaly
	}
	// generate stipple masks
}

/*
 *	bool CGPU_AdaBoost::TClassifierData::operator <(const TClassifierData &r_t_data) const
 *		- less-than operator for weak classifier sorting
 *		- sorts by convolution size, then by block shift
 *		- returns true if this classifier has smaller convolution
 *		  or equal convolution and smaller shift than r_t_data classifier,
 *		  otherwise returns false
 */
bool CGPU_AdaBoost::TClassifierData::operator <(const TClassifierData &r_t_data) const
{
	if(p_classifier->n_block_width == r_t_data.p_classifier->n_block_width &&
	   p_classifier->n_block_height == r_t_data.p_classifier->n_block_height) {
		_ASSERTE(p_convolution == r_t_data.p_convolution);
		// in case block size equals, texture must be identical

		int n_sh0_x = p_classifier->n_position_x % p_classifier->n_block_width;
		int n_sh0_y = p_classifier->n_position_y % p_classifier->n_block_height;
		int n_sh1_x = r_t_data.p_classifier->n_position_x % p_classifier->n_block_width;
		int n_sh1_y = r_t_data.p_classifier->n_position_y % p_classifier->n_block_height;
		// calculate sub-block shift which will affect layer selection

		return n_sh0_x < n_sh1_x || (n_sh0_x == n_sh1_x && n_sh0_y < n_sh1_y);
		// sort by classifier shift (primary key is x, secondary y)
	}

	return p_classifier->n_block_height < r_t_data.p_classifier->n_block_height ||
		(p_classifier->n_block_height == r_t_data.p_classifier->n_block_height &&
		p_classifier->n_block_width < r_t_data.p_classifier->n_block_width);
	// sort by convolution size
}

/*
 *								=== ~CGPU_AdaBoost::TClassifierData ===
 */

/*
 *								=== CGPU_AdaBoost ===
 */

/*
 *	CGPU_AdaBoost::CGPU_AdaBoost(CGLState *p_state, int n_image_width, int n_image_height,
 *		const TAdaBoostClassifier &r_t_classifier, CConvolutionManager &r_conv_plan)
 *		- default constructor
 *		- creates AdaBoost processing plan
 *		- p_state is OpenGL state guard (needed for rasterizer calibration)
 *		- n_image_width and n_image_height are input image width and height
 *		- r_t_classifier contains weak classifiers as loaded from xml
 *		- r_conv_plan is convolution plan (contains convolution images neccessary
 *		  to evaluate for those classifiers)
 *		- if any error occured, classifier list is cleared (number of classifiers returns 0)
 */
CGPU_AdaBoost::CGPU_AdaBoost(CGLState *p_state, int n_image_width, int n_image_height,
	const TAdaBoostClassifier &r_t_classifier, CConvolutionManager &r_conv_plan,
	int n_group_size_limit)
	:m_n_group_size_limit(n_group_size_limit), m_p_frame_buffer(0), m_p_result_texture(0)
{
	m_classifier_data.reserve(r_t_classifier.n_weak_classifier_num);
	if(m_classifier_data.capacity() < r_t_classifier.n_weak_classifier_num)
		return;
	// alloc classifier list

	for(int i = 0; i < r_t_classifier.n_weak_classifier_num; ++ i) {
		m_classifier_data.push_back(TClassifierData(p_state, n_image_width, n_image_height,
			r_t_classifier.p_weak_classifier + i, r_conv_plan));
		if(!m_classifier_data.back().p_Convolution()) {
			m_classifier_data.clear(); // mark error
			return;
		}
	}
	// make list of classifiers along with coresponding convolution textures

	std::sort(m_classifier_data.begin(), m_classifier_data.end());
	// sort by source convolution image

	std::vector<const TClassifierData*> classifier_group_list;
	std::vector<int> classifier_group_first_idx;
	if(!CreateClassifierGroups(classifier_group_list, classifier_group_first_idx))
		return;
	// create classifier groups

	m_runtime_info.reserve(classifier_group_first_idx.size());
	if(m_runtime_info.capacity() < classifier_group_first_idx.size())
		return;
	for(int i = 0, n = classifier_group_first_idx.size(); i < n; ++ i) {
		int n_cls_idx = classifier_group_first_idx[i];
		int n_cls_num = ((i + 1 < n)? classifier_group_first_idx[i + 1] :
			classifier_group_list.size()) - n_cls_idx;

		TClassifierGroupShader cgs(p_state, classifier_group_list.begin() + n_cls_idx,
			classifier_group_list.begin() + n_cls_idx + n_cls_num);
		if(!cgs.p_shader_info[0])
			return;
		// create group shader info

		TClassifierGroupRuntime *p_cgr;
		if(!(p_cgr = new TClassifierGroupRuntime))
			return;
		if(!p_cgr->Create(p_state, cgs)) {
			delete p_cgr;
			return;
		}
		// create new classifier group runtime (invalidates cgs)

		m_runtime_info.push_back(p_cgr);
	}
	// create shaders for high perf frame crunching (in case it fails, m_p_frame_buffer or
	// m_p_result_texture are both 0 and it marks error)

	if(!(m_p_frame_buffer = new CGLRenderBuffer_FBO(n_image_width,
	   n_image_height, 1, true, 0, false, false, 0, false, false, 0)))
		return;
	if(!m_p_frame_buffer->b_Status() || glGetError() != GL_NO_ERROR)
		return;
	// create a new render buffer

	if(!(m_p_result_texture = new CGLTexture_2D(p_state, n_image_width, n_image_height,
	   GL_RGB16F_ARB, false, 0)))
		return;
	// FBO won't render to single-channel texture (todo - see what PBuffer can do or
	// if there's any extension doing that)
	if(!m_p_result_texture->b_Status()) {
		delete m_p_result_texture;
		m_p_result_texture = 0; // mark error
		return;
	}
	// create empty texture to render to
}

/*
 *	void CGPU_AdaBoost::DeleteRTI(TClassifierGroupRuntime *p_cgr)
 *		- cleanup helper
 */
void CGPU_AdaBoost::DeleteRTI(TClassifierGroupRuntime *p_cgr)
{
	delete p_cgr;
}

/*
 *	CGPU_AdaBoost::~CGPU_AdaBoost()
 *		- destructor
 */
CGPU_AdaBoost::~CGPU_AdaBoost()
{
	std::for_each(m_runtime_info.begin(), m_runtime_info.end(), DeleteRTI);
	if(m_p_frame_buffer)
		delete m_p_frame_buffer;
	if(m_p_result_texture)
		delete m_p_result_texture;
}

/*
 *	bool CGPU_AdaBoost::b_Status() const
 *		- returns AdaBoost plan status; if plan is valid returns true, otherwise returns false
 *		- note this reflects constructor outcome
 */
bool CGPU_AdaBoost::b_Status() const
{
	return !m_classifier_data.empty() && m_p_frame_buffer && m_p_result_texture;
}

/*
 *	static int CGPU_AdaBoost::n_Max_VertexInterpolants()
 *		- returns number of vertex interpolants
 *		- OpenGL must be initialized before calling this
 */
int CGPU_AdaBoost::n_Max_VertexInterpolants()
{
	int n_max_texcoord_num;
	glGetIntegerv(GL_MAX_TEXTURE_COORDS, &n_max_texcoord_num);
	return n_max_texcoord_num;
}

/*
 *	static int CGPU_AdaBoost::n_Max_ConvolutionTextures()
 *		- returns number of textures available for convolution textures
 *		- OpenGL must be initialized before calling this
 */
int CGPU_AdaBoost::n_Max_ConvolutionTextures()
{
	int n_max_texture_num;
	glGetIntegerv(GL_MAX_TEXTURE_UNITS, &n_max_texture_num);
	return n_max_texture_num - 1; // one texture is needed for alphas
}

/*
 *	int CGPU_AdaBoost::n_Classifier_Num() const
 *		- returns number of weak classifier
 *		- note this returns 0 if constructor failed
 */
int CGPU_AdaBoost::n_Classifier_Num() const
{
	return m_classifier_data.size();
}

/*
 *	bool CGPU_AdaBoost::ProcessFrame(CGLState *p_state)
 *		- processes a single frame (call convolution manager before calling this)
 *		- p_state is OpenGL state guard
 *		- returns true if everything is ok and no OpenGL errors occur; otherwise returns false
 */
bool CGPU_AdaBoost::ProcessFrame(CGLState *p_state)
{
	if(!b_Status())
		return false;
	// are we ready?

	if(!m_p_frame_buffer->Bind() ||
	   !m_p_frame_buffer->BindColorTexture_2D(*m_p_result_texture, GL_TEXTURE_2D) ||
	   !m_p_frame_buffer->b_Status())
		return false;
	glViewport(0, 0, m_p_frame_buffer->n_Width(), m_p_frame_buffer->n_Height());
	// bind render buffer

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	// reset matrices

	p_state->Enable(GL_POLYGON_STIPPLE);
	// enable stipple mask

	p_state->DisableBlend();
	// disable blending for first frame

	p_state->DisableDepthTest();
	p_state->DisableCullFace();
	// make sure those two are off

	for(int i = 0, n = m_runtime_info.size(); i < n; ++ i) {
		m_runtime_info[i]->Render(p_state);
		if(!i) {
			p_state->BlendFunc(GL_ONE, GL_ONE);
			p_state->EnableBlend();
			// set additive blending after the first frame
		}
	}

	p_state->Disable(GL_POLYGON_STIPPLE);
	// disable stipple mask

	p_state->BindProgramObject_GL2(0);
	// release shader

	if(!m_p_frame_buffer->BindColorTexture_2D(0, GL_TEXTURE_2D) ||
	   !m_p_frame_buffer->Release())
		return false;
	// release render-buffer, put back original viewport

	return glGetError() == GL_NO_ERROR;
}

/*
 *	static bool CGPU_AdaBoost::CompileLinkReport(CGLState *p_state, CGLShaderObject &r_shader)
 *		- compiles shader object r_shader and writes any errors to stderr
 *		- p_state is OpenGL state guard
 *		- returns true on success (shader was compiled and linked), false on failure
 */
bool CGPU_AdaBoost::CompileLinkReport(CGLState *p_state, CGLShaderObject &r_shader)
{
	char *p_s_compile = 0, *p_s_link = 0;
	bool b_result = r_shader.Compile_Link(p_state, p_s_compile, p_s_link);
	if(p_s_compile) {
		printf("\n=== %s compiler report ===\n%s\n",
			r_shader.p_FragmentShader_Info()->p_s_name, p_s_compile);
		delete[] p_s_compile;
	}
	if(p_s_link) {
		printf("\n=== %s linker report ===\n%s\n",
			r_shader.p_FragmentShader_Info()->p_s_name, p_s_link);
		delete[] p_s_link;
	}
	return b_result;
}

/*
 *	static int CGPU_AdaBoost::n_AreaSum(const unsigned char *p_image, int n_width,
 *		int n_x0, int n_y0, int n_x1, int n_y1)
 *		- calculates sum of all pixels of greyscale image p_image which is n_width
 *		  pixels wide under rectangular window n_x0 <= x < n_x1, n_y0 <= y < n_y1
 *		- returns pixel sum
 */
int CGPU_AdaBoost::n_AreaSum(const unsigned char *p_image, int n_width,
	int n_x0, int n_y0, int n_x1, int n_y1)
{
	int n_sum = 0;
	for(; n_x0 < n_x1; ++ n_x0) {
		for(int n_y = n_y0; n_y < n_y1; ++ n_y)
			n_sum += p_image[n_x0 + n_y * n_width];
	}
	return n_sum;
}

bool CGPU_AdaBoost::Validate_ClassifierInputs(CGLState *p_state, CGLTexture_2D *p_source_image,
	CConvolutionManager &r_conv_plan, int &r_n_max_error, float &r_f_std_deviation)
{
	if(!r_conv_plan.Convolve(p_state, p_source_image))
		return false;
	// make sure all convolutions contain correct data

	unsigned char *p_image_buffers;
	int n_image_width = p_source_image->n_Width();
	int n_image_height = p_source_image->n_Height();
	int n_image_size = n_image_width * n_image_height;
	if(!(p_image_buffers = new unsigned char[n_image_size * 2]))
		return false;
	CVectorScopeGuard<unsigned char*> img_del(p_image_buffers);
	// need image buffer for original image (so CPU convolver has a source data),
	// buffer for CPU convolution and for GPU convolution

	unsigned char *p_src_buffer = p_image_buffers;
	unsigned char *p_aux_buffer0 = p_image_buffers + n_image_size;
	// pointers to the image buffers

	glPixelStorei(GL_PACK_ALIGNMENT, 1);
	// byte-aligned pixels

	p_source_image->Bind_Enable(p_state);
	glGetTexImage(GL_TEXTURE_2D, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, p_src_buffer);
	// get source image data

	int n_max_error = 0;
	double f_sum_diff2 = 0;
	unsigned long n_cmp_num = 0;

	const CGLTexture_3D *p_prev_texture = 0;
	for(int i = 0; i < m_classifier_data.size(); ++ i) {
		TClassifierData &r_t_cl = m_classifier_data[i];
		// get some classifier we're going to work on

		if(p_prev_texture != r_t_cl.p_Convolution()) {
			(p_prev_texture = r_t_cl.p_Convolution())->Bind_Enable(p_state);
			glPixelStorei(GL_PACK_ALIGNMENT, 1);
			glGetTexImage(GL_TEXTURE_3D, 0, GL_RGBA, GL_UNSIGNED_BYTE, p_aux_buffer0);
		}
		// get proper texture (as 3D data, though)

		float f_sum_difference2 = 0;
		unsigned long n_comparison_num = 1;

		const int n_x_cutoff = r_t_cl.n_Offset_X() + r_t_cl.n_Grid_Width() * r_t_cl.n_Block_Width();
		const int n_y_cutoff = r_t_cl.n_Offset_Y() + r_t_cl.n_Grid_Height() * r_t_cl.n_Block_Height();
		for(int y = 0; y < n_image_height - n_y_cutoff; ++ y) {
			for(int x = 0; x < n_image_width - n_x_cutoff; ++ x) {
				// x and y are origin of our image window

				int sums[16];
				{
					const int gridSizeX = r_t_cl.n_Grid_Width();
					const int gridSizeY = r_t_cl.n_Grid_Height();
					int *currentSum = sums;
					const int featurePosX = r_t_cl.n_Offset_X();
					const int featurePosY = r_t_cl.n_Offset_Y();
					const int convolutionSizeX = r_t_cl.n_Block_Width();
					const int convolutionSizeY = r_t_cl.n_Block_Height();
					// feed to michal's pseudocode

					for(int yy = 0; yy < gridSizeY; ++ yy) {
						for(int xx = 0; xx < gridSizeX; ++ xx) {
							*currentSum = n_AreaSum(p_src_buffer + x + y * n_image_width,
								n_image_width, featurePosX + xx * convolutionSizeX,
								featurePosY + yy * convolutionSizeY,
								featurePosX + xx * convolutionSizeX + convolutionSizeX,
								featurePosY + yy * convolutionSizeY + convolutionSizeY);

							*currentSum += convolutionSizeX * convolutionSizeY / 2;
							*currentSum /= convolutionSizeX * convolutionSizeY;
							// change to original algorithm - want result in 0-255 range,
							// also round values to get closer to GPU results

							++ currentSum;
						} 
					}
				}
				// calculate gridSizeX per gridSizeY sums, using Michal's reference algorithm

				int my_sums[16];
				{
					const int n_slice_w = r_t_cl.n_Slice_Width();
					const unsigned char *p_slice_buffer = p_aux_buffer0 +
						n_slice_w * r_t_cl.n_Slice_Height() * r_t_cl.n_Slice(x, y);
					// get proper convolution slice

					int x0 = r_t_cl.n_TexCoord_s(x);
					int y0 = r_t_cl.n_TexCoord_t(y);
					int x1 = x0 + r_t_cl.n_Grid_Width();
					int y1 = y0 + r_t_cl.n_Grid_Height();
					// coords of sum rectangle

					int *p_dest_sum = my_sums;
					for(; y0 < y1; ++ y0) {
						for(int xx = x0; xx < x1; ++ xx) {
							*p_dest_sum = p_slice_buffer[xx + y0 * n_slice_w];
							++ p_dest_sum;
						}
					}
					// get my GPU-calculated convolutions
				}

				const int n_grid_w = r_t_cl.n_Grid_Width();
				const int n_grid_h = r_t_cl.n_Grid_Height();
				for(int j = 0; j < n_grid_w * n_grid_h; ++ j) {
					int n_error = abs(sums[j] != my_sums[j]);
					f_sum_difference2 += n_error * n_error;
					if(n_max_error < n_error)
						n_max_error = n_error;
				}
				n_comparison_num += n_grid_w * n_grid_h;
				// compare results
			}
		}
		// check if convolutions are correct

		f_sum_diff2 += f_sum_difference2;
		n_cmp_num += n_comparison_num;
	}

	r_f_std_deviation = (float)sqrt(f_sum_diff2 / n_cmp_num);
	r_n_max_error = n_max_error;

	// cleanup is done automatically :)

	return true;
}

bool CGPU_AdaBoost::Validate_SliceSelection(CGLState *p_state, CGLTexture_2D *p_source_image,
	CConvolutionManager &r_conv_plan, int &r_n_slice_mismatch_num, float &r_f_mismatch_rate)
{
	if(!r_conv_plan.Convolve(p_state, p_source_image))
		return false;
	// make sure all convolutions contain correct data

	unsigned char *p_image_buffers;
	int n_image_width = p_source_image->n_Width();
	int n_image_height = p_source_image->n_Height();
	int n_image_size = n_image_width * n_image_height;
	if(!(p_image_buffers = new unsigned char[n_image_size * 2]))
		return false;
	CVectorScopeGuard<unsigned char*> img_del(p_image_buffers);
	// need image buffer for original image (so CPU convolver has a source data),
	// buffer for CPU convolution and for GPU convolution

	unsigned char *p_src_buffer = p_image_buffers;
	unsigned char *p_aux_buffer0 = p_image_buffers + n_image_size;
	// pointers to the image buffers

	glPixelStorei(GL_PACK_ALIGNMENT, 1);
	// byte-aligned pixels

	p_source_image->Bind_Enable(p_state);
	glGetTexImage(GL_TEXTURE_2D, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, p_src_buffer);
	// get source image data

	CGLRenderBuffer_FBO *p_render_buffer;
	if(!(p_render_buffer = new CGLRenderBuffer_FBO(n_image_width,
	   n_image_height, 1, true, GL_RGBA8, false, false, 0, false, false, 0)))
		return false;
	CScalarScopeGuard<CGLRenderBuffer_FBO*> rb_del(p_render_buffer);
	if(!p_render_buffer->b_Status() || glGetError() != GL_NO_ERROR)
		return false;
	// create a new render buffer

	CGLTexture_2D *p_result_texture;
	if(!(p_result_texture = new CGLTexture_2D(p_state, n_image_width, n_image_height,
	   GL_RGBA8, false, 0)))
		return false;
	CScalarScopeGuard<CGLTexture_2D*> tex_del(p_result_texture);
	if(!p_result_texture->b_Status())
		return false;
	// create empty texture to render to

	TShaderInfo slice_select_shader("slice_select_shader", true,
		"uniform vec4 v_block_to_slice;\n"
		// x and y scaling, y is multiplied by block width. z coord contains half-pixel shift
		"void main()\n"
		"{\n"
		"    vec2 v_slice = v_block_to_slice.xy * gl_TexCoord[0].xy;\n"
		"    v_slice.x = mod(v_slice.x, v_block_to_slice.w);\n"
		"    float f_slice = v_slice.x + v_slice.y + v_block_to_slice.z;\n"
		"    f_slice = mod(f_slice, 1.0);\n" // mod is not necessary when sampling texture
		"    gl_FragColor = vec4(f_slice, f_slice, f_slice, 1);\n"
		"}\n" // calculates 3D texture slice
	);
	CGLShaderObject sh_slice_selection(0, &slice_select_shader);
	if(!CompileLinkReport(p_state, sh_slice_selection))
		return false;
	// create slice selection shader

	p_state->BindProgramObject_GL2(0);
	// make sure no shader is active

	int n_slice_mismatch_num = 0, n_slice_test_num = 0;
	for(int i = 0; i < m_classifier_data.size(); ++ i) {
		TClassifierData &r_t_cl = m_classifier_data[i];
		// get some classifier we're going to work on

		p_state->DisableTexture2D();
		p_state->DisableTexture3D();

		p_render_buffer->Bind();
		p_render_buffer->BindColorTexture_2D(*p_result_texture, GL_TEXTURE_2D);
		glViewport(0, 0, p_render_buffer->n_Width(), p_render_buffer->n_Height());
		// bind render buffer

		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		// reset matrices

		sh_slice_selection.t_Find_Parameter("v_block_to_slice").SetValue(
			r_t_cl.v_block_to_slice_x, r_t_cl.v_block_to_slice_y,
			r_t_cl.v_block_to_slice_z_offset, r_t_cl.v_block_to_slice_x_modulo);
		sh_slice_selection.Bind(p_state);
		// bind slice selection shader (outputs 3d texture slice for each pixel)

		glBegin(GL_QUADS);
		glTexCoord2f(r_t_cl.t_texcoords.f_min_s, r_t_cl.t_texcoords.f_min_t);
		glVertex2f(-1, -1);
		glTexCoord2f(r_t_cl.t_texcoords.f_max_s, r_t_cl.t_texcoords.f_min_t);
		glVertex2f( 1, -1);
		glTexCoord2f(r_t_cl.t_texcoords.f_max_s, r_t_cl.t_texcoords.f_max_t);
		glVertex2f( 1,  1);
		glTexCoord2f(r_t_cl.t_texcoords.f_min_s, r_t_cl.t_texcoords.f_max_t);
		glVertex2f(-1,  1);
		glEnd();
		// render fullscreen quad

		p_state->BindProgramObject_GL2(0);
		// release shader

		p_render_buffer->BindColorTexture_2D(0, GL_TEXTURE_2D);
		p_render_buffer->Release();
		// release render-buffer

		p_result_texture->Bind_Enable(p_state);
		glPixelStorei(GL_PACK_ALIGNMENT, 1);
		glGetTexImage(GL_TEXTURE_2D, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, p_aux_buffer0);

		if(glGetError() != GL_NO_ERROR)
			return false;

		const int n_slice_num = r_t_cl.n_Slice_Num();
		for(int y = 0; y < n_image_height; ++ y) {
			for(int x = 0; x < n_image_width; ++ x) {
				int n_ref_slice = r_t_cl.n_Slice(x, y);
				// calculate image slice no.

				int n_slice = (p_aux_buffer0[x + y * n_image_width] * n_slice_num) / 255;
				// read slice no. generated by shader

				if(n_slice != n_ref_slice)
					++ n_slice_mismatch_num;
				++ n_slice_test_num;
			}
		}
	}
	// determine wheter correct slice is chosen in shader

	r_n_slice_mismatch_num = n_slice_mismatch_num;
	r_f_mismatch_rate = (float)n_slice_mismatch_num / n_slice_test_num;

	// cleanup is done automatically :)

	return true;
}

bool CGPU_AdaBoost::Validate_FeatureSampling(CGLState *p_state, CGLTexture_2D *p_source_image,
	CConvolutionManager &r_conv_plan, bool &r_b_errors)
{
	if(!r_conv_plan.Convolve(p_state, p_source_image))
		return false;
	// make sure all convolutions contain correct data

	unsigned char *p_image_buffers;
	int n_image_width = p_source_image->n_Width();
	int n_image_height = p_source_image->n_Height();
	int n_image_size = n_image_width * n_image_height;
	if(!(p_image_buffers = new unsigned char[n_image_size * 6]))
		return false;
	CVectorScopeGuard<unsigned char*> img_del(p_image_buffers);
	// need image buffer for original image (so CPU convolver has a source data),
	// buffer for CPU convolution and for GPU convolution

	unsigned char *p_src_buffer = p_image_buffers;
	unsigned char *p_aux_buffer0 = p_image_buffers + n_image_size;
	unsigned char *p_aux_buffer1 = p_image_buffers + n_image_size * 2; // RGBA!
	// pointers to the image buffers

	glPixelStorei(GL_PACK_ALIGNMENT, 1);
	// byte-aligned pixels

	p_source_image->Bind_Enable(p_state);
	glGetTexImage(GL_TEXTURE_2D, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, p_src_buffer);
	// get source image data

	CGLRenderBuffer_FBO *p_render_buffer;
	if(!(p_render_buffer = new CGLRenderBuffer_FBO(n_image_width,
	   n_image_height, 1, true, GL_RGBA8, false, false, 0, false, false, 0)))
		return false;
	CScalarScopeGuard<CGLRenderBuffer_FBO*> rb_del(p_render_buffer);
	if(!p_render_buffer->b_Status() || glGetError() != GL_NO_ERROR)
		return false;
	// create a new render buffer

	CGLTexture_2D *p_result_texture;
	if(!(p_result_texture = new CGLTexture_2D(p_state, n_image_width, n_image_height,
	   GL_RGBA8, false, 0)))
		return false;
	CScalarScopeGuard<CGLTexture_2D*> tex_del(p_result_texture);
	if(!p_result_texture->b_Status())
		return false;
	// create empty texture to render to

	TShaderInfo test_sample_shader0("test_sample_shader0", true,
		"uniform sampler3D n_texture;\n"
		"uniform vec4 v_block_to_slice;\n"
		// x and y scaling, y is multiplied by block width. z coord contains half-pixel shift
		"void main()\n"
		"{\n"
		"    vec2 v_slice = v_block_to_slice.xy * gl_TexCoord[0].xy;\n"
		"    v_slice.x = mod(v_slice.x, v_block_to_slice.w);\n"
		"    float f_slice = v_slice.x + v_slice.y + v_block_to_slice.z;\n"
		"    f_slice = mod(f_slice, 1.0);\n" // mod is not necessary when sampling texture
		"    vec3 v_texcoord = vec3(gl_TexCoord[0].xy, f_slice);"
		"    vec3 v_sums0;\n"
		"    v_sums0 = texture3D(n_texture, v_texcoord).xyz;\n"
		"    gl_FragColor = vec4(v_sums0, 1.0);\n"
		"}\n" // calculates 3D texture slice
	);
	CGLShaderObject sh_test_sample0(0, &test_sample_shader0);
	if(!CompileLinkReport(p_state, sh_test_sample0))
		return false;

	TShaderInfo test_sample_shader1("test_sample_shader1", true,
		"uniform sampler3D n_texture;\n"
		"uniform vec4 v_block_to_slice;\n"
		// x and y scaling, y is multiplied by block width. z coord contains half-pixel shift
		"void main()\n"
		"{\n"
		"    vec2 v_slice = v_block_to_slice.xy * gl_TexCoord[0].xy;\n"
		"    v_slice.x = mod(v_slice.x, v_block_to_slice.w);\n"
		"    float f_slice = v_slice.x + v_slice.y + v_block_to_slice.z;\n"
		"    f_slice = mod(f_slice, 1.0);\n" // mod is not necessary when sampling texture
		"    vec3 v_texcoord = vec3(gl_TexCoord[0].xy, f_slice);"
		"    vec3 v_sums0;\n"
		"    v_sums0 = texture3D(n_texture, v_texcoord).yzw;\n"
		"    gl_FragColor = vec4(v_sums0, 1.0);\n"
		"}\n" // calculates 3D texture slice
	);
	CGLShaderObject sh_test_sample1(0, &test_sample_shader1);
	if(!CompileLinkReport(p_state, sh_test_sample1))
		return false;

	TShaderInfo test_sample_shader2("test_sample_shader2", true,
		"uniform sampler3D n_texture;\n"
		"uniform vec4 v_block_to_slice;\n"
		// x and y scaling, y is multiplied by block width. z coord contains half-pixel shift
		"uniform vec2 v_pixel;\n" // "large" 3d-tex slice pixel, not viewport pixel
		"void main()\n"
		"{\n"
		"    vec2 v_slice = v_block_to_slice.xy * gl_TexCoord[0].xy;\n"
		"    v_slice.x = mod(v_slice.x, v_block_to_slice.w);\n"
		"    float f_slice = v_slice.x + v_slice.y + v_block_to_slice.z;\n"
		"    f_slice = mod(f_slice, 1.0);\n" // mod is not necessary when sampling texture
		"    vec3 v_texcoord = vec3(gl_TexCoord[0].xy, f_slice);\n"
		"    vec3 v_sums0;\n"
		"    v_sums0.xy = texture3D(n_texture, v_texcoord).zw;\n"
		"    v_texcoord.x += /*2.0 **/ v_pixel.x;\n"
		// h_ack! intentionally letting there the implicit int to float conversion error
		"    v_sums0.z = texture3D(n_texture, v_texcoord).x;\n"
		"    gl_FragColor = vec4(v_sums0, 1.0);\n"
		"}\n" // calculates 3D texture slice
	);
	CGLShaderObject sh_test_sample2(0, &test_sample_shader2);
	if(!CompileLinkReport(p_state, sh_test_sample2))
		return false;

	TShaderInfo test_sample_shader3("test_sample_shader3", true,
		"uniform sampler3D n_texture;\n"
		"uniform vec4 v_block_to_slice;\n"
		// x and y scaling, y is multiplied by block width. z coord contains half-pixel shift
		"uniform vec2 v_pixel;\n" // "large" 3d-tex slice pixel, not viewport pixel
		"void main()\n"
		"{\n"
		"    vec2 v_slice = v_block_to_slice.xy * gl_TexCoord[0].xy;\n"
		"    v_slice.x = mod(v_slice.x, v_block_to_slice.w);\n"
		"    float f_slice = v_slice.x + v_slice.y + v_block_to_slice.z;\n"
		"    f_slice = mod(f_slice, 1.0);\n" // mod is not necessary when sampling texture
		"    vec3 v_texcoord = vec3(gl_TexCoord[0].xy, f_slice);"
		"    vec3 v_sums0;\n"
		"    v_sums0.x = texture3D(n_texture, v_texcoord).w;\n"
		"    v_texcoord.x += v_pixel.x;\n"
		"    v_sums0.yz = texture3D(n_texture, v_texcoord).xy;\n"
		"    gl_FragColor = vec4(v_sums0, 1.0);\n"
		"}\n" // calculates 3D texture slice
	);
	CGLShaderObject sh_test_sample3(0, &test_sample_shader3);
	if(!CompileLinkReport(p_state, sh_test_sample3))
		return false;
	// create test sampling shaders

	CGLShaderObject *p_test_sample_sh[4] = {
		&sh_test_sample0, &sh_test_sample1, &sh_test_sample2, &sh_test_sample3
	};

	p_state->BindProgramObject_GL2(0);
	// make sure no shader is active

	bool b_feature_sampler_errors = false;
	const CGLTexture_3D *p_prev_texture = 0;
	for(int i = 0; i < m_classifier_data.size(); ++ i) {
		TClassifierData &r_t_cl = m_classifier_data[i];
		// get some classifier we're going to work on

		p_state->DisableTexture2D();
		r_t_cl.p_Convolution()->Bind_Enable(p_state);
		p_state->Color3f(1, 1, 1);

		p_render_buffer->Bind();
		p_render_buffer->BindColorTexture_2D(*p_result_texture, GL_TEXTURE_2D);
		glViewport(0, 0, p_render_buffer->n_Width(), p_render_buffer->n_Height());
		// bind render buffer

		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		// reset matrices

		if(glGetError() != GL_NO_ERROR)
			return false;

		p_state->Enable(GL_POLYGON_STIPPLE);
		// enable stipple mask

		glClear(GL_COLOR_BUFFER_BIT);
		// shouldn't be neccessary ... it's here just for debugging

		for(int n_pass = 0; n_pass < 4; ++ n_pass) {
			glPolygonStipple(r_t_cl.p_Stipple_Mask(n_pass));
			// t_odo - need different stipple/pass patterns for different x-kernel-sizes,
			// investigate into it, design and conquer :) maybe save reference image as
			// well so you don't fix what's not broken

			p_test_sample_sh[n_pass]->t_Find_Parameter("v_block_to_slice").SetValue(
				r_t_cl.v_block_to_slice_x, r_t_cl.v_block_to_slice_y,
				r_t_cl.v_block_to_slice_z_offset, r_t_cl.v_block_to_slice_x_modulo);
			if(n_pass > 1) {
				p_test_sample_sh[n_pass]->t_Find_Parameter("v_pixel").SetValue(r_t_cl.v_pixel_x,
					r_t_cl.v_pixel_y);
			}
			p_test_sample_sh[n_pass]->Bind(p_state);

			glBegin(GL_QUADS);
			glTexCoord2f(r_t_cl.t_texcoords.f_min_s, r_t_cl.t_texcoords.f_min_t);
			glVertex2f(-1, -1);
			glTexCoord2f(r_t_cl.t_texcoords.f_max_s, r_t_cl.t_texcoords.f_min_t);
			glVertex2f( 1, -1);
			glTexCoord2f(r_t_cl.t_texcoords.f_max_s, r_t_cl.t_texcoords.f_max_t);
			glVertex2f( 1,  1);
			glTexCoord2f(r_t_cl.t_texcoords.f_min_s, r_t_cl.t_texcoords.f_max_t);
			glVertex2f(-1,  1);
			glEnd();
			// render fullscreen quad
		}
		// render image in four stippled passes

		p_state->Disable(GL_POLYGON_STIPPLE);
		// disable stipple mask

		p_state->BindProgramObject_GL2(0);
		// release shader

		if(glGetError() != GL_NO_ERROR)
			return false;

		p_render_buffer->BindColorTexture_2D(0, GL_TEXTURE_2D);
		p_render_buffer->Release();
		// release render-buffer, put back original viewport

		p_result_texture->Bind_Enable(p_state);
		glPixelStorei(GL_PACK_ALIGNMENT, 1);
		glGetTexImage(GL_TEXTURE_2D, 0, GL_RGBA, GL_UNSIGNED_BYTE, p_aux_buffer1);
		// download image (RGBA)

		if(glGetError() != GL_NO_ERROR)
			return false;

		if(p_prev_texture != r_t_cl.p_Convolution()) {
			(p_prev_texture = r_t_cl.p_Convolution())->Bind_Enable(p_state);
			glPixelStorei(GL_PACK_ALIGNMENT, 1);
			glGetTexImage(GL_TEXTURE_3D, 0, GL_RGBA, GL_UNSIGNED_BYTE, p_aux_buffer0);
		}
		// get proper texture (as 3D data, though)

		int n_max_error = 0, n_sample_num = 0, p_error_num[3] = {0};
		float f_deviation = 0;

		const int n_x_cutoff = r_t_cl.n_Offset_X() + r_t_cl.n_Grid_Width() * r_t_cl.n_Block_Width();
		const int n_y_cutoff = r_t_cl.n_Offset_Y() + r_t_cl.n_Grid_Height() * r_t_cl.n_Block_Height();
		for(int y = 0; y < n_image_height - n_y_cutoff; ++ y) {
			for(int x = 0; x < n_image_width - n_x_cutoff; ++ x) {
				// x and y are origin of our image window

				int n_slice_w = r_t_cl.n_Slice_Width();
				const unsigned char *p_slice_buffer = p_aux_buffer0 +
					n_slice_w * r_t_cl.n_Slice_Height() * r_t_cl.n_Slice(x, y);

				int x0 = r_t_cl.n_TexCoord_s(x);
				int y0 = r_t_cl.n_TexCoord_t(y);
				int x1 = x0 + r_t_cl.n_Grid_Width();
				int y1 = y0 + r_t_cl.n_Grid_Height();
				// coords of sum rectangle

				int my_sums[16];
				int *p_dest_sum = my_sums;
				for(; y0 < y1; ++ y0) {
					for(int xx = x0; xx < x1; ++ xx) {
						*p_dest_sum = p_slice_buffer[xx + y0 * n_slice_w];
						++ p_dest_sum;
					}
				}
				// get my GPU-assisted sums

				for(int n = 0; n < 3; ++ n) {
					if(p_aux_buffer1[(x + y * n_image_width) * 4 + n] != my_sums[n]) {
						int n_err = abs(p_aux_buffer1[(x + y * n_image_width) * 4 + n] - my_sums[n]);
						if(n_max_error < n_err)
							n_max_error = n_err;
						f_deviation += n_err * n_err;
						++ p_error_num[n];
					}
				}
				// test all three grid components

				n_sample_num += 3;
			}
		}

		f_deviation /= n_sample_num;
		// standard deviation

		if(n_max_error) {
			printf("\twarning: feature sampling %d has std deviation %f (max error %d) (%d %d %d)\n",
				i, f_deviation, n_max_error, p_error_num[0], p_error_num[1], p_error_num[2]);
			b_feature_sampler_errors = true;
			Save_Gray_TGA("first_components.tga", n_image_width * 4, n_image_height, p_aux_buffer1);
		}
	}
	// test feature sampling/swizzling system

	r_b_errors = b_feature_sampler_errors;

	// cleanup is done automatically :)

	return true;
}

bool ShaderSection(int n_grid_width, int n_grid_height, int n_shift, int n_block_a, int n_block_b,
	const std::string &r_s_sampler, const std::string &r_s_texcoord, const std::string &r_s_pixel,
	const std::string &r_s_block_to_slice, const std::string &r_s_alpha_sampler,
	const std::string &r_s_alpha_pixel_size, int n_half_alpha_num,
	const std::string &r_s_alpha_swizzle, const std::string &r_s_alpha_t, std::string &r_s_result)
{
	if(n_grid_width < 2 || n_grid_width > 4)
		return false;
	// refuse different block sizes

	r_s_result  = "{\n";
	r_s_result += "    vec3 v_texcoord;\n";
	r_s_result += "    v_texcoord.yz = " + r_s_block_to_slice + ".xy * " + r_s_texcoord + ";\n";
	r_s_result += "    v_texcoord.y = mod(v_texcoord.y, " + r_s_block_to_slice + ".w);\n";
	r_s_result += "    v_texcoord.z += v_texcoord.y + " + r_s_block_to_slice + ".z;\n";
	// v_texcoord.z contains (unclamped) texture slice. now ...

	r_s_result += "    v_texcoord.xy = " + r_s_texcoord + ";\n";
	// v_texcoord.xy contains appropriate pixel coordinates, texcoords are complete

	bool b_double_fetch;
	if(b_double_fetch = (n_grid_width + n_shift > 4)) {
		r_s_result += "    vec3 v_texcoord2 = v_texcoord;\n";
		r_s_result += "    v_texcoord2.x += " + r_s_pixel + ".x;\n";
		// add second texcoord register for the texel to the right
	}

	char p_s_sums_type[5];
	sprintf(p_s_sums_type, "vec%d", n_grid_width);
	// one line of sums will be stored in vec2, vec3 or vec4

	r_s_result += std::string("    ") + p_s_sums_type + " ";
	for(int i = 0; i < n_grid_height; ++ i) {
		char p_s_var_name[8];
		sprintf(p_s_var_name, "v_sums%d", i);
		// variable name for this line of sums

		if(i)
			r_s_result += ", ";
		r_s_result += p_s_var_name;
	}
	r_s_result += ";\n";
	// variable definitions

	for(int i = 0; i < n_grid_height; ++ i) {
		char p_s_var_name[8];
		sprintf(p_s_var_name, "v_sums%d", i);
		// variable name for this line of sums

		// 00001111 <- smp
		// xyzwxyzw <- swz
		// 01234567 <- off

		int n_length = min(4 - n_shift, n_grid_width);
		// length of first swizzle

		char p_s_swizzle_w[5];
		strncpy(p_s_swizzle_w, "xyzw", n_length);
		p_s_swizzle_w[n_length] = 0;
		// write mask

		char p_s_swizzle_r[5];
		strncpy(p_s_swizzle_r, "xyzwxyzwxyzw" + n_shift, n_length);
		p_s_swizzle_r[n_length] = 0;
		// swizzle

		r_s_result += "    " + (p_s_var_name + ("." + (p_s_swizzle_w +
			(" = texture3D(" + r_s_sampler)))) + ", v_texcoord)." + p_s_swizzle_r + ";\n";
		// create command (the parentheses enforces such execution order to avoid pointer addition)

		if(b_double_fetch) {
			int n_length1 = max(0, n_grid_width - n_length);
			// length of second swizzle

			char p_s_swizzle_w[5];
			strncpy(p_s_swizzle_w, "xyzw" + n_length, n_length1);
			p_s_swizzle_w[n_length1] = 0;
			// write mask

			char p_s_swizzle_r[5];
			strncpy(p_s_swizzle_r, "xyzwxyzwxyzw" + n_length + n_shift, n_length1);
			p_s_swizzle_r[n_length1] = 0;
			// swizzle

			r_s_result += "    " + (p_s_var_name + ("." + (p_s_swizzle_w +
				(" = texture3D(" + r_s_sampler)))) + ", v_texcoord2)." + p_s_swizzle_r + ";\n";
			// create command
		}

		r_s_result += "    v_texcoord.y += " + r_s_pixel + ".y;\n";
		if(b_double_fetch)
			r_s_result += "    v_texcoord2.y += " + r_s_pixel + ".y;\n";
		// move texcoords to another line
	}
	// creates block that fills the grid of n_grid_width per n_grid_height samples

	char p_s_block_a[10];
	sprintf(p_s_block_a, "v_sums%d.%c",
		n_block_a / n_grid_width, n_block_a % n_grid_width + 'x');
	char p_s_block_b[10];
	sprintf(p_s_block_b, "v_sums%d.%c",
		n_block_b / n_grid_width, n_block_b % n_grid_width + 'x');
	// create names of block a and b
	// (for example 3x3 grid, n_block_a = 3 -> p_s_block_a = "v_sums1.x")

	r_s_result += std::string("    i") + p_s_sums_type + " v_cmp_accum;\n";
	// declare ivec variable for accumulating comparison results

	for(int i = 0; i < n_grid_height; ++ i) {
		char p_s_var_name[8];
		sprintf(p_s_var_name, "v_sums%d", i);
		// variable name for this line of sums

		r_s_result += std::string((i)? "    v_cmp_accum += i" : "    v_cmp_accum = i") +
			p_s_sums_type + "(greaterThan(" + p_s_sums_type + "(" + p_s_block_a +
			"), " + p_s_var_name + "));\n";
		r_s_result += std::string("    v_cmp_accum -= i") + p_s_sums_type +
			"(greaterThan(" + p_s_sums_type + "(" + p_s_block_b + "), " + p_s_var_name + "));\n";
		// sum up comparison results, converted to ivec (fixme - ivec is signed, right?)
	}

	r_s_result += std::string("    float f_rank_diff = dot(") + p_s_sums_type + "(1.0), " +
		p_s_sums_type + "(v_cmp_accum));\n"; // + " + p_s_offset + ".0;\n";
	// calculate rank difference (offset is realised by wrapping texture arround)

	/*
	char p_s_offset[16];
	sprintf(p_s_offset, "%d", n_half_alpha_num);

	r_s_result += ("    f_result += texture2D(" + r_s_alpha) + "[int(f_rank_diff) +
		p_s_offset + "]." + r_s_alpha_swizzle + ";\n";*/
	// alphas in register array (not in this shader profile :-/)

	r_s_result += "    f_result += texture2D(" + r_s_alpha_sampler +
		", vec2(.5 + f_rank_diff, " + r_s_alpha_t + ") * " + r_s_alpha_pixel_size +
		")." + r_s_alpha_swizzle + ";\n";
	// alphas in texture

	// generate final code that sums up comparisons (using dot with 1 to sum elements,
	// comparisons are carried out in vectors) and indexes the alpha array

	r_s_result += "}\n";

	return true;
}

std::string s_Indent(std::string s_source_code)
{
	s_source_code.insert(0, "    ");
	for(int n = s_source_code.find('\n', 0);
	   n != std::string::npos; n = s_source_code.find('\n', n)) {
		if(++ n >= s_source_code.length())
			break; // do not indent last line (which is empty btw)
		if(s_source_code[n] == '\n')
			continue; // do not indent empty lines
		s_source_code.insert(n, "    ");
		n += 4;
	}
	return s_source_code;
}

int n_GEqual_Pot(int n_size, int n_max = unsigned int(-1) / 2) // returns -1 on failure
{
	int n_result = 1;
	while(n_result < n_size) {
		if(n_result >= n_max / 2)
			return -1;
		n_result *= 2;
	}
	return n_result;
}

bool CreateTextureList(std::vector<const CGPU_AdaBoost::TClassifierData*>::iterator p_begin,
	std::vector<const CGPU_AdaBoost::TClassifierData*>::iterator p_end,
	std::vector<const CGLTexture_3D*> &r_texture_list)
{
	r_texture_list.clear();
	// make sure it's empty

	for(; p_begin != p_end; ++ p_begin) {
		const CGLTexture_3D *p_cur_tex = (*p_begin)->p_Convolution();
		std::vector<const CGLTexture_3D*>::iterator p_texture_iter;
		if((p_texture_iter = std::find(r_texture_list.begin(),
		   r_texture_list.end(), p_cur_tex)) == r_texture_list.end()) {
			r_texture_list.reserve(r_texture_list.size() + 1);
			if(r_texture_list.capacity() == r_texture_list.size())
				return false;
			r_texture_list.push_back(p_cur_tex);
		}
	}
	// create list of used textures

	return true;
}

bool CGPU_AdaBoost::CreateClassifierGroups(
	std::vector<const TClassifierData*> &r_classifier_group_list,
	std::vector<int> &r_classifier_group_first_idx)
{
	const int n_max_texcoord_num = n_Max_VertexInterpolants();
	const int n_max_texture_num = n_Max_ConvolutionTextures();

	for(int i = 0; i < m_classifier_data.size(); ++ i) {
		const TClassifierData &r_t_cl = m_classifier_data[i];
		// get some classifier we're going to work on

		int n_group = -1;
		for(int j = 0, n = r_classifier_group_first_idx.size(); j < n; ++ j) {
			int n_cls_idx = r_classifier_group_first_idx[j];
			int n_cls_num = ((j + 1 < n)? r_classifier_group_first_idx[j + 1] :
				r_classifier_group_list.size()) - n_cls_idx;

			if(n_cls_num == m_n_group_size_limit)
				continue;
			// this one is full

			const TClassifierData &r_t_head = *r_classifier_group_list[n_cls_idx];
			// get first classifier in group

			if(r_t_cl.b_Compatible(r_t_head)) {
				int n_cls_idx = r_classifier_group_first_idx[j];
				int n_cls_num = ((j + 1 < n)? r_classifier_group_first_idx[j + 1] :
					r_classifier_group_list.size()) - n_cls_idx;

				std::vector<const CGLTexture_3D*> texture_list;
				if(!CreateTextureList(r_classifier_group_list.begin() + n_cls_idx,
				   r_classifier_group_list.begin() + n_cls_idx + n_cls_num, texture_list))
					return false;
				// create texture list

				const CGLTexture_3D *p_ref_tex = r_t_cl.p_Convolution();

				texture_list.reserve(texture_list.size() + 1);
				if(texture_list.capacity() == texture_list.size())
					return false;
				texture_list.push_back(p_ref_tex);
				// insert sentinel

				if((std::find(texture_list.begin(), texture_list.end(),
				   p_ref_tex) - texture_list.begin()) >= n_max_texture_num)
					continue;
				// see wheter textures fit to available texturing units

				std::vector<CRasterizerCalibrator::TTexCoordRect> texcoord_list;
				for(int k = n_cls_idx, n = n_cls_idx + n_cls_num; k < n; ++ k) {
					const TClassifierData *p_cl = r_classifier_group_list[k];
					if(std::find(texcoord_list.begin(), texcoord_list.end(),
					   p_cl->t_texcoords) == texcoord_list.end()) {
						texcoord_list.reserve(texcoord_list.size() + 1);
						if(texcoord_list.capacity() == texcoord_list.size())
							return false;
						texcoord_list.push_back(p_cl->t_texcoords);
					}
				}
				// create list of texcoords

				texcoord_list.reserve(texcoord_list.size() + 1);
				if(texcoord_list.capacity() == texcoord_list.size())
					return false;
				texcoord_list.push_back(r_t_cl.t_texcoords);
				// insert sentinel

				if((std::find(texcoord_list.begin(), texcoord_list.end(),
				   r_t_cl.t_texcoords) - texcoord_list.begin()) >= n_max_texcoord_num)
					continue;
				// see wheter texcoords fit to available vertex attributes

				n_group = j;
				break;
			}
			// can it be in the same group?
		}
		// try to find group where r_t_cl fits

		r_classifier_group_list.reserve(r_classifier_group_list.size() + 1);
		if(r_classifier_group_list.capacity() == r_classifier_group_list.size())
			return false;

		if(n_group == -1) {
			r_classifier_group_first_idx.reserve(r_classifier_group_first_idx.size() + 1);
			if(r_classifier_group_first_idx.capacity() == r_classifier_group_first_idx.size())
				return false;

			r_classifier_group_first_idx.push_back(r_classifier_group_list.size());
			r_classifier_group_list.push_back(&r_t_cl);
			// add new group
		} else {
			if(n_group + 1 < r_classifier_group_first_idx.size()) {
				r_classifier_group_list.insert(r_classifier_group_list.begin() +
					r_classifier_group_first_idx[n_group + 1], &r_t_cl);
				// in case weren't inserting to last group, add the classifier
				// to the end of it (ie that's where the next one starts)

				for(int j = n_group + 1; j < r_classifier_group_first_idx.size(); ++ j)
					++ r_classifier_group_first_idx[j];
				// offset beginning indices of all following groups
			} else {
				r_classifier_group_list.push_back(&r_t_cl);
				// otherwise just stash it in the back
			}
		}
		// manage double list (that's because i didn't want list of lists)
	}

	return true;
}

bool CGPU_AdaBoost::Validate_ShaderSynthesis(CGLState *p_state, CGLTexture_2D *p_source_image,
	CConvolutionManager &r_conv_plan, bool &r_b_errors)
{
	if(!r_conv_plan.Convolve(p_state, p_source_image))
		return false;
	// make sure all convolutions contain correct data

	unsigned char *p_image_buffers;
	int n_image_width = p_source_image->n_Width();
	int n_image_height = p_source_image->n_Height();
	int n_image_size = n_image_width * n_image_height;
	if(!(p_image_buffers = new unsigned char[n_image_size * 2 + 2 * n_image_size * sizeof(float)]))
		return false;
	CVectorScopeGuard<unsigned char*> img_del(p_image_buffers);
	// need image buffer for original image (so CPU convolver has a source data),
	// buffer for CPU convolution and for GPU convolution

	unsigned char *p_src_buffer = p_image_buffers;
	unsigned char *p_aux_buffer0 = p_image_buffers + n_image_size;
	float *p_aux_buffer1 = (float*)(p_image_buffers + n_image_size * 2); // hack! ubyte -> float
	float *p_aux_buffer2 = p_aux_buffer1 + n_image_size;
	// pointers to the image buffers

	glPixelStorei(GL_PACK_ALIGNMENT, 1);
	// byte-aligned pixels

	p_source_image->Bind_Enable(p_state);
	glGetTexImage(GL_TEXTURE_2D, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, p_src_buffer);
	// get source image data

	const GLenum n_alpha_fmt = GL_ALPHA16F_ARB;
	const GLenum n_result_internal_fmt = GL_RGB16F_ARB; // FBO won't render to single-channel texture (todo - see what PBuffer can do or if there's any extension doing that)
	const GLenum n_result_fmt = GL_RED; // get red channel

	CGLRenderBuffer_FBO *p_render_buffer;
	if(!(p_render_buffer = new CGLRenderBuffer_FBO(n_image_width,
	   n_image_height, 1, true, 0, false, false, 0, false, false, 0)))
		return false;
	CScalarScopeGuard<CGLRenderBuffer_FBO*> rb_del(p_render_buffer);
	if(!p_render_buffer->b_Status() || glGetError() != GL_NO_ERROR)
		return false;
	// create a new render buffer

	CGLTexture_2D *p_result_texture;
	if(!(p_result_texture = new CGLTexture_2D(p_state, n_image_width, n_image_height,
	   n_result_internal_fmt, false, 0))) // debug, as with render-buffer
		return false;
	CScalarScopeGuard<CGLTexture_2D*> tex_del(p_result_texture);
	if(!p_result_texture->b_Status())
		return false;
	// create empty texture to render to

	if(glGetError() != GL_NO_ERROR)
		return false;

	/*printf("\tthis may take a while ...\n");

	bool b_feature_sampler_errors = false;
	const CGLTexture_3D *p_prev_texture = 0;
	for(int i = 0; i < m_classifier_data.size(); ++ i) {
		printf("\tcompiling shaders for classifier %d ...\r", i + 1);

		TClassifierData &r_t_cl = m_classifier_data[i];
		// get some classifier we're going to work on

		p_state->DisableTexture2D();
		r_t_cl.p_Convolution()->Bind_Enable(p_state);
		p_state->Color3f(1, 1, 1);

		if(!p_render_buffer->Bind() ||
		   !p_render_buffer->BindColorTexture_2D(*p_result_texture, GL_TEXTURE_2D))
		    fprintf(stderr, "error: OpenGL error(s) detected\n");
		glViewport(0, 0, p_render_buffer->n_Width(), p_render_buffer->n_Height());
		// bind render buffer

		if(!p_render_buffer->b_Status()) {
			int n = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
		    fprintf(stderr, "error: FBO error(s) detected (0x%04X)\n", n);
		}

		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		// reset matrices

		if(glGetError() != GL_NO_ERROR)
			return false;

		p_state->Enable(GL_POLYGON_STIPPLE);
		// enable stipple mask

		//glClear(GL_COLOR_BUFFER_BIT);
		// shouldn't be neccessary ... it's here just for debugging

		int n_err;
		if((n_err = glGetError()) != GL_NO_ERROR)
			fprintf(stderr, "error: uncaught OpenGL error(s) pending (0x%04X)\n", n_err);

		for(int n_pass = 0; n_pass < 4; ++ n_pass) {
			glPolygonStipple(r_t_cl.p_Stipple_Mask(n_pass));
			// t_odo - need different stipple/pass patterns for different x-kernel-sizes,
			// investigate into it, design and conquer :) maybe save reference image as
			// well so you don't fix what's not broken

			std::string s_shader;
			if(!ShaderSection(r_t_cl.n_Grid_Width(), r_t_cl.n_Grid_Height(), n_pass, r_t_cl.n_Block_A(), r_t_cl.n_Block_B(),
			   std::string("n_texture"), std::string("gl_TexCoord[0].xy"), std::string("v_pixel"), std::string("v_block_to_slice"),
			   std::string("n_alphas"), std::string("v_alpha_pixel"), r_t_cl.n_Alpha_Num() / 2, std::string("w"), std::string("0.5"), s_shader)) {
				p_state->Disable(GL_POLYGON_STIPPLE);
				p_state->BindProgramObject_GL2(0);
				return false;
			}
			// create shader string

			const char *p_s_vars =
				"uniform sampler3D n_texture;\n"
				"uniform sampler2D n_alphas;\n"
				"uniform vec2 v_alpha_pixel;\n" // size of pixel in alpha texture
				"uniform vec4 v_block_to_slice;\n"
				"uniform vec2 v_pixel;\n"; // "large" 3d-tex slice pixel, not viewport pixel
			// shader variable block (t_odo - write alpha texture management)

			const char *p_s_prolog =
				"void main()\n"
				"{\n"
				"    float f_result = 0.0;\n\n";
			// prolog where main starts

			const char *p_s_epilog =
				"\n    gl_FragColor.x = f_result;\n"
				"}\n";
			// epilog where result is written to framebuffer and we're done

			int n_err;
			if((n_err = glGetError()) != GL_NO_ERROR)
				fprintf(stderr, "error: uncaught OpenGL error(s) pending (0x%04X)\n", n_err);

			TShaderInfo ab_shader("synth_ab_shader", true,
				(p_s_vars + (p_s_prolog + (s_Indent(s_shader) + p_s_epilog))).c_str(), "n_texture|n_alphas");

			//printf("%s\n", ab_shader.p_high->p_s_source_code);

			CGLShaderObject sh_aboost(0, &ab_shader);
			if(!CompileLinkReport(p_state, sh_aboost)) {
				p_state->Disable(GL_POLYGON_STIPPLE);
				p_state->BindProgramObject_GL2(0);
				return false;
			}
			// compile the shader

			int n_alpha_size = n_GEqual_Pot(r_t_cl.n_Alpha_Num());
			if(n_alpha_size == -1) {
				p_state->Disable(GL_POLYGON_STIPPLE);
				p_state->BindProgramObject_GL2(0);
				return false;
			}
			// calculate alpha texture size (nearest greater power of two)

			if(r_t_cl.n_Alpha_Num() > 256) {
				p_state->Disable(GL_POLYGON_STIPPLE);
				p_state->BindProgramObject_GL2(0);
				return false;
			}
			float p_alpha[256] = {0};
			int n_offset = r_t_cl.n_Alpha_Num() / 2;
			for(int j = 0; j < r_t_cl.n_Alpha_Num(); ++ j)
				p_alpha[(j - n_offset + n_alpha_size) % n_alpha_size] = r_t_cl.f_Alpha(j);
			// fill the alpha array, realise offset here

			p_state->ActiveTextureUnit(1);

			CGLTexture_2D alpha_tex(p_state, n_alpha_size, 1,
				n_alpha_fmt, false, 0, GL_ALPHA, GL_FLOAT, p_alpha);
			if(!alpha_tex.b_Status()) {
				p_state->Disable(GL_POLYGON_STIPPLE);
				p_state->BindProgramObject_GL2(0);
				p_state->ActiveTextureUnit(0);
				return false;
			}
			// create (1D) texture with alphas, values are oriented horizontally

			alpha_tex.Bind_Enable(p_state);
			alpha_tex.r_Parameters(p_state).Set_Texture_Wrap_S(GL_REPEAT); // this is how we handle param offset! don't clamp.
			alpha_tex.r_Parameters(p_state).Set_Texture_Wrap_T(GL_CLAMP);
			alpha_tex.r_Parameters(p_state).Set_Texture_Mag_Filter(GL_NEAREST);
			alpha_tex.r_Parameters(p_state).Set_Texture_Min_Filter(GL_NEAREST);
			// no filtering & clamp to edge

			//int n_err;
			if((n_err = glGetError()) != GL_NO_ERROR)
				fprintf(stderr, "error: uncaught OpenGL error(s) pending (0x%04X)\n", n_err);

			float f_alpha_pixel = 1.0f / n_alpha_size;

			sh_aboost.t_Find_Parameter("v_block_to_slice").SetValue(
				r_t_cl.v_block_to_slice_x, r_t_cl.v_block_to_slice_y,
				r_t_cl.v_block_to_slice_z_offset, r_t_cl.v_block_to_slice_x_modulo);
			sh_aboost.t_Find_Parameter("v_pixel").SetValue(r_t_cl.v_pixel_x,
				r_t_cl.v_pixel_y);
			sh_aboost.t_Find_Parameter("v_alpha_pixel").SetValue(f_alpha_pixel, 1);
			//sh_aboost.t_Find_Parameter("p_alpha").SetValue(p_alpha, 0, r_t_cl.n_Alpha_Num());
			sh_aboost.Bind(p_state);
			// bind the shader

			glBegin(GL_QUADS);
			glTexCoord2f(r_t_cl.t_texcoords.f_min_s, r_t_cl.t_texcoords.f_min_t);
			glVertex2f(-1, -1);
			glTexCoord2f(r_t_cl.t_texcoords.f_max_s, r_t_cl.t_texcoords.f_min_t);
			glVertex2f( 1, -1);
			glTexCoord2f(r_t_cl.t_texcoords.f_max_s, r_t_cl.t_texcoords.f_max_t);
			glVertex2f( 1,  1);
			glTexCoord2f(r_t_cl.t_texcoords.f_min_s, r_t_cl.t_texcoords.f_max_t);
			glVertex2f(-1,  1);
			glEnd();
			// render fullscreen quad

			alpha_tex.Disable(p_state);
			p_state->ActiveTextureUnit(0);
		}
		// render image in four stippled passes

		if(glGetError() != GL_NO_ERROR)
			return false;

		p_state->Disable(GL_POLYGON_STIPPLE);
		// disable stipple mask

		p_state->BindProgramObject_GL2(0);
		// release shader

		if(glGetError() != GL_NO_ERROR)
			return false;

		p_render_buffer->BindColorTexture_2D(0, GL_TEXTURE_2D);
		p_render_buffer->Release();
		// release render-buffer, put back original viewport

		p_result_texture->Bind_Enable(p_state);
		glPixelStorei(GL_PACK_ALIGNMENT, 1);
		glGetTexImage(GL_TEXTURE_2D, 0, n_result_fmt, GL_FLOAT, p_aux_buffer1);
		// download image (float luminance)

		if(glGetError() != GL_NO_ERROR)
			return false;

		if(p_prev_texture != r_t_cl.p_Convolution()) {
			(p_prev_texture = r_t_cl.p_Convolution())->Bind_Enable(p_state);
			glPixelStorei(GL_PACK_ALIGNMENT, 1);
			glGetTexImage(GL_TEXTURE_3D, 0, GL_RGBA, GL_UNSIGNED_BYTE, p_aux_buffer0);
		}
		// get proper texture (as 3D data, though)

		const int n_x_cutoff = r_t_cl.n_Offset_X() + r_t_cl.n_Grid_Width() * r_t_cl.n_Block_Width();
		const int n_y_cutoff = r_t_cl.n_Offset_Y() + r_t_cl.n_Grid_Height() * r_t_cl.n_Block_Height();
		for(int y = 0; y < n_image_height - n_y_cutoff; ++ y) {
			for(int x = 0; x < n_image_width - n_x_cutoff; ++ x) {
				// x and y are origin of our image window

				int n_slice_w = r_t_cl.n_Slice_Width();
				const unsigned char *p_slice_buffer = p_aux_buffer0 +
					n_slice_w * r_t_cl.n_Slice_Height() * r_t_cl.n_Slice(x, y);

				int x0 = r_t_cl.n_TexCoord_s(x);
				int y0 = r_t_cl.n_TexCoord_t(y);
				int x1 = x0 + r_t_cl.n_Grid_Width();
				int y1 = y0 + r_t_cl.n_Grid_Height();
				// coords of sum rectangle

				int my_sums[16];
				int *p_dest_sum = my_sums;
				for(; y0 < y1; ++ y0) {
					for(int xx = x0; xx < x1; ++ xx) {
						*p_dest_sum = p_slice_buffer[xx + y0 * n_slice_w];
						++ p_dest_sum;
					}
				}
				// get my GPU-assisted sums

				int n_rank_diff = r_t_cl.n_Alpha_Num() / 2;
				// fixme - classifier array offset

				for(int i = 0, n = r_t_cl.n_Grid_Width() * r_t_cl.n_Grid_Height(); i < n; ++ i) {
					n_rank_diff += my_sums[r_t_cl.n_Block_A()] > my_sums[i];
					n_rank_diff -= my_sums[r_t_cl.n_Block_B()] > my_sums[i];
				}
				// rank calculation

				float f_result = r_t_cl.f_Alpha(n_rank_diff);
				// fetch final alpha

				if(fabs(p_aux_buffer1[x + y * n_image_width] - f_result) > .001f)
					b_feature_sampler_errors = true;
				// todo - do more comperhensive error checking
			}
		}

		if(b_feature_sampler_errors)
			break;
	}
	// test feature sampling/swizzling system

	printf("\n\tdone\n");*/

	std::vector<const TClassifierData*> classifier_group_list;
	std::vector<int> classifier_group_first_idx;
	// classifier groups

	const int n_max_texcoord_num = n_Max_VertexInterpolants();
	const int n_max_texture_num = n_Max_ConvolutionTextures();

	printf("\tthere are %d free texturing units and %d texcoords\n",
		n_max_texture_num, n_max_texcoord_num);

	printf("\tgroup size limit is %d classifiers (24 was ok with 7950GT)\n",
		m_n_group_size_limit);

	if(!CreateClassifierGroups(classifier_group_list, classifier_group_first_idx))
		return false;
	// create classifier groups

	bool b_feature_sampler_errors = false;

	printf("\tthere are %d classifier groups by mask and texture matching\n", classifier_group_first_idx.size());
	for(int i = 0, n = classifier_group_first_idx.size(); i < n; ++ i) {
		int n_cls_idx = classifier_group_first_idx[i];
		int n_cls_num = ((i + 1 < n)? classifier_group_first_idx[i + 1] :
			classifier_group_list.size()) - n_cls_idx;

		std::vector<const CGLTexture_3D*> texture_list;
		if(!CreateTextureList(classifier_group_list.begin() + n_cls_idx,
		   classifier_group_list.begin() + n_cls_idx + n_cls_num, texture_list))
			return false;
		// create texture list

		printf("\thave %d classifiers using %d convolution textures in group %d. probing ...\n",
			n_cls_num, texture_list.size(), i);

		TClassifierGroupShader sh(p_state, classifier_group_list.begin() + n_cls_idx,
			classifier_group_list.begin() + n_cls_idx + n_cls_num);
		// create shader info's

		CGLShaderObject *p_shader[4] = {0};
		for(int j = 0; j < 4; ++ j) {
			if(!sh.p_shader_info[j]) {
				fprintf(stderr, "\terror: failed to generate shader"
					" source codes (not enough memory)\n");
				for(int k = 0; k < j; ++ k)
					delete p_shader[k];
				return false;
			}
			if(!(p_shader[j] = new CGLShaderObject(0, sh.p_shader_info[j]))) {
				fprintf(stderr, "\terror: not enough memory\n");
				for(int k = 0; k < j; ++ k)
					delete p_shader[k];
				return false;
			}
			if(!CompileLinkReport(p_state, *p_shader[j])) {
				for(int k = 0; k <= j; ++ k)
					delete p_shader[k];
				fprintf(stderr, "\terror: failed to comlipe shaders."
					" try smaller group size limit\n\t       if you believe"
					" it's going to remedy error(s) you see above\n");
				return false;
			}
		}
		// try to compile that (omfg, omfg)

		bool b_result = TestShaderRun(p_state, sh, p_shader, p_result_texture, p_render_buffer,
			p_aux_buffer0, p_aux_buffer1, p_aux_buffer2);
		// test shader run

		for(int j = 0; j < 4; ++ j)
			delete p_shader[j];
		// delete shaders

		if(!b_result)
			return false;
	}
	// print classifier groups

	r_b_errors = b_feature_sampler_errors;

	// cleanup is done automatically :)

	return true;
}

bool CGPU_AdaBoost::TestShaderRun(CGLState *p_state,
	const TClassifierGroupShader &r_sh, CGLShaderObject *p_shader[4],
	CGLTexture_2D *p_result_texture, CGLRenderBuffer_FBO *p_render_buffer,
	unsigned char *p_aux_buffer0, float *p_aux_buffer1, float *p_aux_buffer2)
{
	for(int i = 0, n = r_sh.texture_list.size(); i < n; ++ i) {
		p_state->ActiveTextureUnit(i);
		p_state->DisableTexture2D();
		r_sh.texture_list[i]->Bind_Enable(p_state);
	}
	// bind convolution textures

	p_state->ActiveTextureUnit(r_sh.texture_list.size());
	p_state->DisableTexture3D();
	r_sh.p_alpha_texture->Bind_Enable(p_state);
	// bind alpha texture

	int n_texture_unit_num;
	glGetIntegerv(GL_MAX_TEXTURE_UNITS, &n_texture_unit_num);
	for(int i = r_sh.texture_list.size(); i < n_texture_unit_num; ++ i) {
		p_state->ActiveTextureUnit(i);
		p_state->DisableTexture2D();
		p_state->DisableTexture3D();
	}
	// disable additional textures

	p_state->ActiveTextureUnit(0);
	p_state->Color3f(1, 1, 1);

	p_render_buffer->Bind();
	p_render_buffer->BindColorTexture_2D(*p_result_texture, GL_TEXTURE_2D);
	glViewport(0, 0, p_render_buffer->n_Width(), p_render_buffer->n_Height());
	// bind render buffer

	if(!p_render_buffer->b_Status()) {
		int n = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
		fprintf(stderr, "error: FBO error(s) detected (0x%04X)\n", n);
		return false;
	}

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	// reset matrices

	p_state->Enable(GL_POLYGON_STIPPLE);
	// enable stipple mask

	{int n_err;
	if((n_err = glGetError()) != GL_NO_ERROR) {
		fprintf(stderr, "error: uncaught OpenGL error(s) pending (0x%04X)\n", n_err);
		return false;
	}}

	/*CTimer t;
	glFlush();
	glFinish();
	float f_start = t.f_Time();
	const int n_test_runs = 1000;
	for(int x = 0; x < n_test_runs; ++ x) {*/
	 // benchmark

	for(int n_pass = 0; n_pass < 4; ++ n_pass) {
		glPolygonStipple(r_sh.classifier_list[0]->p_Stipple_Mask(n_pass));
		// set polygon stipple for this pass

		CGLShaderObject &sh_aboost = *p_shader[n_pass];
		// we're going to use this shader

		for(int i = 0, n = r_sh.block_to_slice_list.size(); i < n; ++ i) {
			vec4 v_b2s = r_sh.block_to_slice_list[i];
			sh_aboost.t_Find_Parameter(r_sh.p_s_BlocktoSlice_Name(i)).SetValue(
				v_b2s.x, v_b2s.y, v_b2s.z, v_b2s.w);
		}
		for(int i = 0, n = r_sh.pixel_list.size(); i < n; ++ i) {
			vec2 v_pixel_size = r_sh.pixel_list[i];
			sh_aboost.t_Find_Parameter(r_sh.p_s_Pixel_Name(i)).SetValue(
				v_pixel_size.x, v_pixel_size.y);
		}
		sh_aboost.t_Find_Parameter("v_alpha_pixel").SetValue(
			r_sh.v_alpha_pixel.x, r_sh.v_alpha_pixel.y);
		// set shader parameters (hopefully that's all)

		sh_aboost.Bind(p_state);
		// bind the shader

		glBegin(GL_QUADS); // t_odo - add multitexcoords list to r_sh
		for(int i = 0, n = r_sh.texcoord_list.size(); i < n; ++ i) {
			glMultiTexCoord2f(GL_TEXTURE0 + i,
				r_sh.texcoord_list[i].f_min_s, r_sh.texcoord_list[i].f_min_t);
		}
		glVertex2f(-1, -1);
		for(int i = 0, n = r_sh.texcoord_list.size(); i < n; ++ i) {
			glMultiTexCoord2f(GL_TEXTURE0 + i,
				r_sh.texcoord_list[i].f_max_s, r_sh.texcoord_list[i].f_min_t);
		}
		glVertex2f( 1, -1);
		for(int i = 0, n = r_sh.texcoord_list.size(); i < n; ++ i) {
			glMultiTexCoord2f(GL_TEXTURE0 + i,
				r_sh.texcoord_list[i].f_max_s, r_sh.texcoord_list[i].f_max_t);
		}
		glVertex2f( 1,  1);
		for(int i = 0, n = r_sh.texcoord_list.size(); i < n; ++ i) {
			glMultiTexCoord2f(GL_TEXTURE0 + i,
				r_sh.texcoord_list[i].f_min_s, r_sh.texcoord_list[i].f_max_t);
		}
		glVertex2f(-1,  1);
		glEnd();
		// render fullscreen quad
	}
	// render image in four stippled passes

	/*}
	glFlush();
	glFinish();
	float f_end = t.f_Time();
	printf("\tbenchmark: %g ms per frame\n", (f_end - f_start) / n_test_runs * 1000);*/
	// ~benchmark

	p_state->Disable(GL_POLYGON_STIPPLE);
	// disable stipple mask

	p_state->BindProgramObject_GL2(0);
	// release shader

	p_render_buffer->BindColorTexture_2D(0, GL_TEXTURE_2D);
	p_render_buffer->Release();
	// release render-buffer, put back original viewport

	{int n_err;
	if((n_err = glGetError()) != GL_NO_ERROR) {
		fprintf(stderr, "error: uncaught OpenGL error(s) pending (0x%04X)\n", n_err);
		return false;
	}}

	const GLenum n_result_fmt = GL_RED; // get red channel

	p_result_texture->Bind_Enable(p_state);
	glPixelStorei(GL_PACK_ALIGNMENT, 1);
	glGetTexImage(GL_TEXTURE_2D, 0, n_result_fmt, GL_FLOAT, p_aux_buffer1);
	// download GPU results

	{int n_err;
	if((n_err = glGetError()) != GL_NO_ERROR) {
		fprintf(stderr, "error: uncaught OpenGL error(s) pending (0x%04X)\n", n_err);
		return false;
	}}

	int n_image_width = p_render_buffer->n_Width();
	int n_image_height = p_render_buffer->n_Height();

	memset(p_aux_buffer2, 0, n_image_width * n_image_height * sizeof(float));
	// clear CPU results buffer

	int n_max_x_cutoff = 0;
	int n_max_y_cutoff = 0;

	const CGLTexture_3D *p_prev_texture = 0;
	for(int i = 0, n = r_sh.classifier_list.size(); i < n; ++ i) {
		const TClassifierData &r_t_cl = *r_sh.classifier_list[i];

		if(p_prev_texture != r_t_cl.p_Convolution()) {
			(p_prev_texture = r_t_cl.p_Convolution())->Bind_Enable(p_state);
			glPixelStorei(GL_PACK_ALIGNMENT, 1);
			glGetTexImage(GL_TEXTURE_3D, 0, GL_RGBA, GL_UNSIGNED_BYTE, p_aux_buffer0);
		}
		// get proper texture (as 3D data, though)

		const int n_x_cutoff = r_t_cl.n_Offset_X() + r_t_cl.n_Grid_Width() * r_t_cl.n_Block_Width();
		const int n_y_cutoff = r_t_cl.n_Offset_Y() + r_t_cl.n_Grid_Height() * r_t_cl.n_Block_Height();

		if(n_max_x_cutoff < n_x_cutoff)
			n_max_x_cutoff = n_x_cutoff;
		if(n_max_y_cutoff < n_y_cutoff)
			n_max_y_cutoff = n_y_cutoff;

		for(int y = 0; y < n_image_height - n_y_cutoff; ++ y) {
			for(int x = 0; x < n_image_width - n_x_cutoff; ++ x) {
				// x and y are origin of our image window

				int n_slice_w = r_t_cl.n_Slice_Width();
				const unsigned char *p_slice_buffer = p_aux_buffer0 +
					n_slice_w * r_t_cl.n_Slice_Height() * r_t_cl.n_Slice(x, y);

				int x0 = r_t_cl.n_TexCoord_s(x);
				int y0 = r_t_cl.n_TexCoord_t(y);
				int x1 = x0 + r_t_cl.n_Grid_Width();
				int y1 = y0 + r_t_cl.n_Grid_Height();
				// coords of sum rectangle

				int my_sums[16];
				int *p_dest_sum = my_sums;
				for(; y0 < y1; ++ y0) {
					for(int xx = x0; xx < x1; ++ xx) {
						*p_dest_sum = p_slice_buffer[xx + y0 * n_slice_w];
						++ p_dest_sum;
					}
				}
				// get my GPU-assisted sums

				int n_rank_diff = r_t_cl.n_Alpha_Num() / 2;
				// fixme - classifier array offset

				for(int i = 0, n = r_t_cl.n_Grid_Width() * r_t_cl.n_Grid_Height(); i < n; ++ i) {
					n_rank_diff += my_sums[r_t_cl.n_Block_A()] > my_sums[i];
					n_rank_diff -= my_sums[r_t_cl.n_Block_B()] > my_sums[i];
				}
				// rank calculation

				float f_result = r_t_cl.f_Alpha(n_rank_diff);
				// fetch final alpha

				p_aux_buffer2[x + y * n_image_width] += f_result;
				// accumulate results to buffer
			}
		}
	}
	// calculate results for the same set of classifiers on CPU

	for(int y = 0; y < n_image_height - n_max_y_cutoff; ++ y) {
		for(int x = 0; x < n_image_width - n_max_x_cutoff; ++ x) {
			if(fabs(p_aux_buffer1[x + y * n_image_width] -
			   p_aux_buffer2[x + y * n_image_width]) > .01f) {
				printf("\t... and you're in trouble\n");
				return false;
			}
		}
	}
	// compare results (just look for differences)

	return true;
}

/*
 *								=== ~CGPU_AdaBoost ===
 */

/*
 *								=== CGPU_AdaBoost::TClassifierGroupShader ===
 */

CGPU_AdaBoost::TClassifierGroupShader::TClassifierGroupShader(CGLState *p_state,
	TClassifierIterator p_begin, TClassifierIterator p_end)
	:p_alpha_texture(0)
{
	for(int i = 0; i < 4; ++ i)
		p_shader_info[i] = 0;

	classifier_list.reserve(p_end - p_begin);
	if(classifier_list.capacity() < p_end - p_begin)
		return;
	classifier_list.insert(classifier_list.begin(), p_begin, p_end);
	// copy classifier list

	if(!Create(p_state)) {
		for(int i = 0; i < 4; ++ i) {
			if(p_shader_info[i]) {
				delete p_shader_info[i];
				p_shader_info[i] = 0;
			}
		}
		// make sure shader info's are all zero on failure
	}
}

CGPU_AdaBoost::TClassifierGroupShader::~TClassifierGroupShader()
{
	if(p_alpha_texture)
		delete p_alpha_texture;
	for(int i = 0; i < 4; ++ i) {
		if(p_shader_info[i])
			delete p_shader_info[i];
	}
}

bool CGPU_AdaBoost::TClassifierGroupShader::Create(CGLState *p_state)
{
#ifdef DUMP_SHADER_SOURCES
	static int n_cls_min = 0;
	int n_cls_max = n_cls_min + classifier_list.size();
	// debug
#endif //DUMP_SHADER_SOURCES

	if(!CreateTextureList(classifier_list.begin(), classifier_list.end(), texture_list))
		return false;
	// create list of used textures

	int n_max = 1;
	for(int i = 0, n = classifier_list.size(); i < n; ++ i) {
		if(classifier_list[i]->n_Alpha_Num() > n_max)
			n_max = classifier_list[i]->n_Alpha_Num();
	}
	int n_max_tex_size;
	glGetIntegerv(GL_MAX_TEXTURE_SIZE, &n_max_tex_size);
	// find maximum of alphas

	int n_alpha_texture_width = n_GEqual_Pot(n_max, n_max_tex_size);
	int n_alpha_texture_height = n_GEqual_Pot(classifier_list.size(), n_max_tex_size);
	if(n_alpha_texture_width == -1 || n_alpha_texture_height == -1)
		return false; // won't fit
	// calculate alpha texture dimensions

	v_alpha_pixel.x = 1.0f / n_alpha_texture_width;
	v_alpha_pixel.y = 1.0f / n_alpha_texture_height;
	// size of alpha texture pixel in OpenGL

	float *p_alpha;
	if(!(p_alpha = new float[n_alpha_texture_width * n_alpha_texture_height]))
		return false;
	memset(p_alpha, 0, n_alpha_texture_width * n_alpha_texture_height * sizeof(float));
	// alloc data for alpha texture

	for(int i = 0, n = classifier_list.size(); i < n; ++ i) {
		float *p_alpha_scanline = p_alpha + i * n_alpha_texture_width;
		// dest scanline

		const TClassifierData *p_cl = classifier_list[i];
		int n_offset = p_cl->n_Alpha_Num() / 2;
		for(int j = 0; j < p_cl->n_Alpha_Num(); ++ j) {
			p_alpha_scanline[(j - n_offset + n_alpha_texture_width) %
				n_alpha_texture_width] = p_cl->f_Alpha(j);
		}
		// fill the alpha array, realise offset here
	}

	if(!(p_alpha_texture = new CGLTexture_2D(p_state, n_alpha_texture_width,
	   n_alpha_texture_height, GL_ALPHA16F_ARB, false, 0, GL_ALPHA, GL_FLOAT, p_alpha))) {
		delete[] p_alpha;
		return false;
	}
	// create alpha texture

	p_alpha_texture->Bind_Enable(p_state);
	p_alpha_texture->r_Parameters(p_state).Set_Texture_Wrap_S(GL_REPEAT); // this is how we handle param offset! don't clamp.
	p_alpha_texture->r_Parameters(p_state).Set_Texture_Wrap_T(GL_CLAMP);
	p_alpha_texture->r_Parameters(p_state).Set_Texture_Mag_Filter(GL_NEAREST);
	p_alpha_texture->r_Parameters(p_state).Set_Texture_Min_Filter(GL_NEAREST);
	// no filtering & clamp to edge

	delete[] p_alpha;
	// don't need this anymore

	for(int i = 0, n = classifier_list.size(); i < n; ++ i) {
		const TClassifierData *p_cl = classifier_list[i];
		vec4 v_block_to_slice(p_cl->v_block_to_slice_x,
			p_cl->v_block_to_slice_y, p_cl->v_block_to_slice_z_offset,
			p_cl->v_block_to_slice_x_modulo);
		if(std::find(block_to_slice_list.begin(), block_to_slice_list.end(),
		   v_block_to_slice) == block_to_slice_list.end()) {
			block_to_slice_list.reserve(block_to_slice_list.size() + 1);
			if(block_to_slice_list.capacity() == block_to_slice_list.size())
				return false;
			block_to_slice_list.push_back(v_block_to_slice);
		}
	}
	// create list of v_block_to_slice vectors

	for(int i = 0, n = classifier_list.size(); i < n; ++ i) {
		const TClassifierData *p_cl = classifier_list[i];
		if(std::find(texcoord_list.begin(), texcoord_list.end(),
		   p_cl->t_texcoords) == texcoord_list.end()) {
			texcoord_list.reserve(texcoord_list.size() + 1);
			if(texcoord_list.capacity() == texcoord_list.size())
				return false;
			texcoord_list.push_back(p_cl->t_texcoords);
		}
	}
	// create list of texcoords

	if(texcoord_list.size() > 8)
		fprintf(stderr, "\twarning: %d texcoords\n", texcoord_list.size());

	for(int i = 0, n = classifier_list.size(); i < n; ++ i) {
		const TClassifierData *p_cl = classifier_list[i];
		vec2 v_pixel(p_cl->v_pixel_x, p_cl->v_pixel_y);
		if(std::find(pixel_list.begin(), pixel_list.end(),
		   v_pixel) == pixel_list.end()) {
			pixel_list.reserve(pixel_list.size() + 1);
			if(pixel_list.capacity() == pixel_list.size())
				return false;
			pixel_list.push_back(v_pixel);
		}
	}
	// create list of v_pixel vectors

	std::string s_vars;
	std::string s_samplers;

	{
		s_vars += "uniform sampler3D ";
		for(int i = 0, n = texture_list.size(); i < n; ++ i) {
			const char *p_s_name = p_s_Texture_Name(i);

			s_vars += p_s_name;
			s_vars += (i + 1 < n)? ", " : ";\n";

			s_samplers += p_s_name;
			s_samplers += "|";
		}
		// convolution texture samplers

		s_samplers += "n_alphas";

		s_vars += "uniform vec2 ";
		for(int i = 0, n = pixel_list.size(); i < n; ++ i) {
			s_vars += p_s_Pixel_Name(i);
			s_vars += (i + 1 < n)? ", " : ";\n";
		}
		// it's pixel size vectors

		s_vars += "uniform sampler2D n_alphas;\n";
		s_vars += "uniform vec2 v_alpha_pixel;\n";
		// alpha texture & it's pixel size

		s_vars += "uniform vec4 ";
		for(int i = 0, n = block_to_slice_list.size(); i < n; ++ i) {
			s_vars += p_s_BlocktoSlice_Name(i);
			s_vars += (i + 1 < n)? ", " : ";\n";
		}
		// block_to_slice vectors
	}
	// create shader variables

	for(int n_pass = 0; n_pass < 4; ++ n_pass) {
		std::string s_body;

		const TClassifierData *p_ref_cl = classifier_list[0];
		for(int i = 0, n = classifier_list.size(); i < n; ++ i) {
			const TClassifierData *p_cl = classifier_list[i];

			int n_texture_index = std::find(texture_list.begin(), texture_list.end(),
				p_cl->p_Convolution()) - texture_list.begin();
			const char *p_s_texture_name = p_s_Texture_Name(n_texture_index);
			// create convolution texture name

			int n_texcoord_index = std::find(texcoord_list.begin(), texcoord_list.end(),
				p_cl->t_texcoords) - texcoord_list.begin();
			const char *p_s_texcoords_name = p_s_TexCoords_Name(n_texcoord_index);
			// create convolution texture name

			vec2 v_blah(p_cl->v_pixel_x, p_cl->v_pixel_y);
			int n_pixel_index = std::find(pixel_list.begin(), pixel_list.end(),
				v_blah) - pixel_list.begin();
			const char *p_s_pixel_name = p_s_Pixel_Name(n_pixel_index);
			// create v_pixel name

			vec4 v_block_to_slice(p_cl->v_block_to_slice_x,
				p_cl->v_block_to_slice_y, p_cl->v_block_to_slice_z_offset,
				p_cl->v_block_to_slice_x_modulo);
			int n_block_to_slice_index = std::find(block_to_slice_list.begin(),
				block_to_slice_list.end(), v_block_to_slice) - block_to_slice_list.begin();
			const char *p_s_block_to_slice_name = p_s_BlocktoSlice_Name(n_block_to_slice_index);
			// create v_block_to_slice name

			float f_alpha_t = i + .5f;
			char p_s_alpha_t[64];
			sprintf(p_s_alpha_t, "%f", f_alpha_t);
			// t coordinate in alpha texture (it is supposed to be upside down, right?)

			int n_phase = (p_ref_cl->n_Phase(*p_cl) + 4 + n_pass) % 4;
			// select classifier pass

			std::string s_shader_section;
			if(!ShaderSection(p_cl->n_Grid_Width(), p_cl->n_Grid_Height(), n_phase,
			   p_cl->n_Block_A(), p_cl->n_Block_B(), std::string(p_s_texture_name),
			   std::string(p_s_texcoords_name), std::string(p_s_pixel_name),
			   std::string(p_s_block_to_slice_name), std::string("n_alphas"),
			   std::string("v_alpha_pixel"), p_cl->n_Alpha_Num() / 2, std::string("w"),
			   std::string(p_s_alpha_t), s_shader_section))
				return false;
			// create shader section

			char p_s_comment[128];
			sprintf(p_s_comment, "// shader section for #%d, phase %d\n\n", i, n_phase);
			// add comment

			s_body += s_shader_section;
			s_body += p_s_comment;
		}
		// create shader sections

		std::string s_main;

		s_main = s_vars + "\n";
		s_main += "void main()\n"
				  "{\n"
				  "    float f_result = .0;\n\n";
		s_main += s_Indent(s_body);
		s_main += "    gl_FragColor.x = f_result;\n"
				  "}\n";
		// concat code to create full shader body for this pass

		if(!(p_shader_info[n_pass] = new TShaderInfo("gpu_aboost",
		   true, s_main.c_str(), s_samplers.c_str())) || !p_shader_info[n_pass]->p_high) {
			return false;
		}
		// create shader source

#ifdef DUMP_SHADER_SOURCES
		char p_s_filename[128];
		sprintf(p_s_filename, "shader_c%d-%d_p%d.glsl.txt", n_cls_min, n_cls_max, n_pass);
		FILE *p_fw = fopen(p_s_filename, "w");
		if(p_fw) {
			fprintf(p_fw, "%s", s_main.c_str());
			fclose(p_fw);
		}
		// write to file (debug)
#endif //DUMP_SHADER_SOURCES
	}

#ifdef DUMP_SHADER_SOURCES
	n_cls_min = n_cls_max;
	// debug
#endif //DUMP_SHADER_SOURCES

	return true;
}

const char *CGPU_AdaBoost::TClassifierGroupShader::p_s_BlocktoSlice_Name(int n_index)
{
	static char p_s_block_to_slice_name[64];
	sprintf(p_s_block_to_slice_name, "v_block_to_slice_%02d", n_index);
	return p_s_block_to_slice_name;
}

const char *CGPU_AdaBoost::TClassifierGroupShader::p_s_Pixel_Name(int n_index)
{
	static char p_s_pixel_name[64];
	sprintf(p_s_pixel_name, "v_pixel_%02d", n_index);
	return p_s_pixel_name;
}

const char *CGPU_AdaBoost::TClassifierGroupShader::p_s_Texture_Name(int n_index)
{
	static char p_s_texture_name[64];
	sprintf(p_s_texture_name, "n_texture_%d", n_index);
	return p_s_texture_name;
}

const char *CGPU_AdaBoost::TClassifierGroupShader::p_s_TexCoords_Name(int n_index)
{
	static char p_s_texcoords_name[64];
	sprintf(p_s_texcoords_name, "gl_TexCoord[%d].xy", n_index);
	return p_s_texcoords_name;
}

/*
 *								=== ~CGPU_AdaBoost::TClassifierGroupShader ===
 */

/*
 *								=== CGPU_AdaBoost::TClassifierGroupRuntime ===
 */

CGPU_AdaBoost::TClassifierGroupRuntime::TClassifierGroupRuntime()
	:p_alpha_texture(0), n_display_list(0)
{
	for(int i = 0; i < 4; ++ i)
		p_shader[i] = 0;
}

CGPU_AdaBoost::TClassifierGroupRuntime::~TClassifierGroupRuntime()
{
	for(int i = 0; i < 4; ++ i) {
		if(p_shader[i])
			delete p_shader[i];
	}
	if(n_display_list)
		glDeleteLists(n_display_list, 1);
}

bool CGPU_AdaBoost::TClassifierGroupRuntime::Create(CGLState *p_state,
	TClassifierGroupShader &r_cgs)
{
	texture_list.swap(r_cgs.texture_list);
	if(!(p_alpha_texture = r_cgs.p_alpha_texture))
		return false;
	r_cgs.p_alpha_texture = 0; // don't want it to be deleted
	// swap textures with cgs

	for(int i = 0; i < 4; ++ i) {
		memcpy(p_stipple_mask[i], r_cgs.classifier_list[0]->p_Stipple_Mask(i),
			32 * 4 * sizeof(unsigned char));
	}
	// copy stipple masks

	for(int j = 0; j < 4; ++ j) {
		if(!r_cgs.p_shader_info[j])
			return false;
		if(!(p_shader[j] = new CGLShaderObject(0, r_cgs.p_shader_info[j]))) {
			for(int k = 0; k < j; ++ k)
				delete p_shader[k];
			return false;
		}
		char *p_s_compile = 0, *p_s_link = 0;
		bool b_result = p_shader[j]->Compile_Link(p_state, p_s_compile, p_s_link);
		if(p_s_compile)
			delete[] p_s_compile;
		if(p_s_link)
			delete[] p_s_link;
		if(!b_result) {
			for(int k = 0; k <= j; ++ k)
				delete p_shader[k];
			return false;
		}
		// create and compile shader

		CGLShaderObject &sh_aboost = *p_shader[j];
		for(int i = 0, n = r_cgs.block_to_slice_list.size(); i < n; ++ i) {
			vec4 v_b2s = r_cgs.block_to_slice_list[i];
			if(!sh_aboost.t_Find_Parameter(r_cgs.p_s_BlocktoSlice_Name(i)).SetValue(
			   v_b2s.x, v_b2s.y, v_b2s.z, v_b2s.w))
				return false;
		}
		for(int i = 0, n = r_cgs.pixel_list.size(); i < n; ++ i) {
			vec2 v_pixel_size = r_cgs.pixel_list[i];
			if(!sh_aboost.t_Find_Parameter(r_cgs.p_s_Pixel_Name(i)).SetValue(
			   v_pixel_size.x, v_pixel_size.y))
				return false;
		}
		if(!sh_aboost.t_Find_Parameter("v_alpha_pixel").SetValue(
		   r_cgs.v_alpha_pixel.x, r_cgs.v_alpha_pixel.y))
			return false;
		// set shader parameters
	}
	// alloc & compile shaders, set params

	if(!(n_display_list = glGenLists(1)))
		return false;
	glNewList(n_display_list, GL_COMPILE);

	glBegin(GL_QUADS); // t_odo - add multitexcoords list to r_sh
	for(int i = 0, n = r_cgs.texcoord_list.size(); i < n; ++ i) {
		glMultiTexCoord2f(GL_TEXTURE0 + i,
			r_cgs.texcoord_list[i].f_min_s, r_cgs.texcoord_list[i].f_min_t);
	}
	glVertex2f(-1, -1);
	for(int i = 0, n = r_cgs.texcoord_list.size(); i < n; ++ i) {
		glMultiTexCoord2f(GL_TEXTURE0 + i,
			r_cgs.texcoord_list[i].f_max_s, r_cgs.texcoord_list[i].f_min_t);
	}
	glVertex2f( 1, -1);
	for(int i = 0, n = r_cgs.texcoord_list.size(); i < n; ++ i) {
		glMultiTexCoord2f(GL_TEXTURE0 + i,
			r_cgs.texcoord_list[i].f_max_s, r_cgs.texcoord_list[i].f_max_t);
	}
	glVertex2f( 1,  1);
	for(int i = 0, n = r_cgs.texcoord_list.size(); i < n; ++ i) {
		glMultiTexCoord2f(GL_TEXTURE0 + i,
			r_cgs.texcoord_list[i].f_min_s, r_cgs.texcoord_list[i].f_max_t);
	}
	glVertex2f(-1,  1);
	glEnd();

	glEndList();
	// create fullscreen quad submission display list

	return glGetError() == GL_NO_ERROR;
}

void CGPU_AdaBoost::TClassifierGroupRuntime::Render(CGLState *p_state) const
{
	for(int i = 0, n = texture_list.size(); i < n; ++ i) {
		p_state->ActiveTextureUnit(i);
		p_state->DisableTexture2D();
		texture_list[i]->Bind_Enable(p_state);
	}
	// bind convolution textures

	p_state->ActiveTextureUnit(texture_list.size());
	p_state->DisableTexture3D();
	p_alpha_texture->Bind_Enable(p_state);
	// bind alpha texture

	// no need to disable / enable other textures,
	// we're using shader so they shouldn't trouble us

	p_state->Enable(GL_POLYGON_STIPPLE); // enabled by CGPU_AdaBoost::ProcessFrame()
	// enable stipple mask

	for(int n_pass = 0; n_pass < 4; ++ n_pass) {
		glPolygonStipple(p_stipple_mask[n_pass]);
		// set polygon stipple for this pass

		p_shader[n_pass]->Bind(p_state);
		// bind the shader (parameters are set and they never change)

		glCallList(n_display_list);
		// render fullscreen quad
	}
}

/*
 *								=== ~CGPU_AdaBoost::TClassifierGroupRuntime ===
 */
