/*
								+-----------------------------------+
								|                                   |
								|  ***  LRD classifier engine  ***  |
								|                                   |
								|   Copyright   -tHE SWINe- 2007   |
								|                                   |
								|           Classifier.h            |
								|                                   |
								+-----------------------------------+
*/

#ifndef __LRD_CLASSIFIER_GPU_ENGINE_INCLUDED
#define __LRD_CLASSIFIER_GPU_ENGINE_INCLUDED

/*

  group classifiers by source convolutions

  create vertex shader that would calculate proper source texcoords for classifiers

  create fragment shader that would group 


-------------------------------------------------------

  classifier results for different pixel offsets come out in different textures
  need to interleave the values; do it:

  using branching (determine pixel offset, use branch that will calculate proper result) - slow

  fill 4 results at once (RGBA pack) - unable to use early alpha test for cutoff, max convolution size will be 4 (4 offsets)

  render to multiple buffers at once - have to interleave results later which could be problematic, requires extra pass,
	limited to 4 texture units anyway

  decision: fill 4 results at once; in case convolutions are larger, use branching

  !!!!! need to use branching anyway, RGBA pack helps with horizontal offsets only

  have different offsets of convolution in different layers of 3d texture

  it's terrible mess, all this ... but on the other hand texture accesses necessary to read 3x3 block are minimized

*/

/*
 *	class CGPU_AdaBoost
 *		- GPU AdaBoost engine class (creates operation plan
 *		  (once) and then executes it (many times))
 */
class CGPU_AdaBoost {
protected:
	/*
	 *	struct CGPU_AdaBoost::TClassifierData
	 *		- data for weak classifier execution
	 *		  (source texture, shader parameters, stipple mask)
	 */
	struct TClassifierData {
	protected:
		const TWeakClassifier *p_classifier;
		// classifier parameters from xml file

		const CGLTexture_3D *p_convolution;
		// convolution image (owned by CConvolutionManager)

		unsigned char p_stipple_mask[4][32 * 4];
		// stipple masks

	public:
		CRasterizerCalibrator::TTexCoordRect t_texcoords;
		// calibrated texcoords for shader

		float v_block_to_slice_x;
		float v_block_to_slice_x_modulo;
		float v_block_to_slice_y;
		float v_block_to_slice_z_offset;
		// slice selection shader parameters

		float v_pixel_x, v_pixel_y;
		// 3d texture texel dimensions

		/*
		 *	CGPU_AdaBoost::TClassifierData::TClassifierData(CGLState *p_state,
		 *		int n_image_width, int n_image_height, const TWeakClassifier *_p_classifier,
		 *		CConvolutionManager &r_conv_plan)
		 *		- default constructor
		 *		- p_state is OpenGL state guard (needed for rasterizer calibration)
		 *		- n_image_width and n_image_height are input image width and height
		 *		- _p_classifier contains weak classifier parameters as loaded from xml
		 *		- r_conv_plan is convolution plan (contains convolution image neccessary
		 *		  to evaluate for this classifier)
		 *		- if any error occured, convolution image is set to 0 (p_Convolution() returns 0)
		 */
		TClassifierData(CGLState *p_state, int n_image_width, int n_image_height,
			const TWeakClassifier *_p_classifier, CConvolutionManager &r_conv_plan);

		/*
		 *	bool CGPU_AdaBoost::TClassifierData::operator <(const TClassifierData &r_t_data) const
		 *		- less-than operator for weak classifier sorting
		 *		- sorts by convolution size, then by block shift
		 *		- returns true if this classifier has smaller convolution
		 *		  or equal convolution and smaller shift than r_t_data classifier,
		 *		  otherwise returns false
		 */
		bool operator <(const TClassifierData &r_t_data) const;

		/*
		 *	inline int CGPU_AdaBoost::TClassifierData::n_Slice_Num() const
		 *		- returns number of convlolution texture slices (each one contains convolution
		 *		  with different image shift), also equals to convolution size (kernel area)
		 */
		inline int n_Slice_Num() const
		{
			return p_classifier->n_block_width * p_classifier->n_block_height;
		}

		/*
		 *	inline int CGPU_AdaBoost::TClassifierData::n_Block_Width() const
		 *		- returns width of convolution kernel
		 */
		inline int n_Block_Width() const
		{
			return p_classifier->n_block_width;
		}

		/*
		 *	inline int CGPU_AdaBoost::TClassifierData::n_Block_Height() const
		 *		- returns height of convolution kernel
		 */
		inline int n_Block_Height() const
		{
			return p_classifier->n_block_height;
		}

		/*
		 *	inline int CGPU_AdaBoost::TClassifierData::n_Slice_Width() const
		 *		- returns width of convolved image in samples.
		 *		  that's four times more width in pixels (RGBA packing)
		 */
		inline int n_Slice_Width() const
		{
			return p_convolution->n_Width() * 4;
		}

		/*
		 *	inline int CGPU_AdaBoost::TClassifierData::n_Slice_Height() const
		 *		- returns height of convolved image (in samples or pixels, no packing here)
		 */
		inline int n_Slice_Height() const
		{
			return p_convolution->n_Height();
		}

		/*
		 *	inline int CGPU_AdaBoost::TClassifierData::n_Grid_Width() const
		 *		- returns feature grid width (number of feature horizontal convolutions)
		 */
		inline int n_Grid_Width() const
		{
			return p_classifier->n_width;
		}

		/*
		 *	inline int CGPU_AdaBoost::TClassifierData::n_Grid_Height() const
		 *		- returns feature grid height (number of feature vertical convolutions)
		 */
		inline int n_Grid_Height() const
		{
			return p_classifier->n_height;
		}

		/*
		 *	inline int CGPU_AdaBoost::TClassifierData::n_Offset_X() const
		 *		- returns horizontal offset of first sample from window origin in pixels
		 */
		inline int n_Offset_X() const
		{
			return p_classifier->n_position_x;
		}

		/*
		 *	inline int CGPU_AdaBoost::TClassifierData::n_Offset_Y() const
		 *		- returns vertical offset of first sample from window origin in pixels
		 */
		inline int n_Offset_Y() const
		{
			return p_classifier->n_position_y;
		}

		/*
		 *	inline int CGPU_AdaBoost::TClassifierData::n_Alpha_Num() const
		 *		- returns number of alphas
		 */
		inline int n_Alpha_Num() const
		{
			return p_classifier->n_alpha_num;
		}

		/*
		 *	inline float CGPU_AdaBoost::TClassifierData::f_Alpha(int n_index) const
		 *		- returns alpha coefficient with zero-based index n_index
		 */
		inline float f_Alpha(int n_index) const
		{
			return p_classifier->p_alpha[n_index];
		}

		/*
		 *	inline const unsigned char *CGPU_AdaBoost::TClassifierData::p_Stipple_Mask(
		 *		int n_pass) const
		 *		- returns stipple mask for pass n_pass (0 to 3)
		 */
		inline const unsigned char *p_Stipple_Mask(int n_pass) const
		{
			return p_stipple_mask[n_pass];
		}

		/*
		 *	const CGLTexture_3D *p_Convolution() const
		 *		- returns convolution texture for this classifier or 0 on error
		 */
		const CGLTexture_3D *p_Convolution() const
		{
			return p_convolution;
		}

		/*
		 *	inline int CGPU_AdaBoost::TClassifierData::n_Block_A() const
		 *		- returns zero-based index of the first feature block to be compared with others
		 */
		inline int n_Block_A() const
		{
			return p_classifier->n_block_a;
		}

		/*
		 *	inline int CGPU_AdaBoost::TClassifierData::n_Block_B() const
		 *		- returns zero-based index of second feature block to be compared with others
		 */
		inline int n_Block_B() const
		{
			return p_classifier->n_block_b;
		}

		// todo - add custom fuckin' designed shader for each classifier ... save merging for later

		/*
		 *	inline int CGPU_AdaBoost::TClassifierData::n_Slice(int x, int y) const
		 *		- returns zero-based index of texture slice for x and y in output image
		 */
		inline int n_Slice(int x, int y) const
		{
			return (x + p_classifier->n_position_x) % p_classifier->n_block_width +
				((y + p_classifier->n_position_y) % p_classifier->n_block_height) *
				p_classifier->n_block_width;
		}

		/*
		 *	inline int CGPU_AdaBoost::TClassifierData::n_TexCoord_s(int x) const
		 *		- returns texture coordinate s for x position in output image (in pixels)
		 */
		inline int n_TexCoord_s(int x) const
		{
			return (x + p_classifier->n_position_x) / p_classifier->n_block_width;
		}

		/*
		 *	inline int CGPU_AdaBoost::TClassifierData::n_TexCoord_t(int y) const
		 *		- returns texture coordinate t for y position in output image (in pixels)
		 */
		inline int n_TexCoord_t(int y) const
		{
			return (y + p_classifier->n_position_y) / p_classifier->n_block_height;
		}

		/*
		 *	inline bool CGPU_AdaBoost::TClassifierData::b_Compatible(
		 *		const TClassifierData &r_t_classifier) const
		 *		- returns true if stipple mask of this classifier and r_t_classifier
		 *		  are the same masks with possible phase shift, otherwise returns false
		 */
		inline bool b_Compatible(const TClassifierData &r_t_classifier) const
		{
			return p_classifier->n_block_width == r_t_classifier.p_classifier->n_block_width &&
				p_classifier->n_block_width != 0 &&
				p_classifier->n_position_x % p_classifier->n_block_width ==
				r_t_classifier.p_classifier->n_position_x % p_classifier->n_block_width;
			// n_block_width != 0 is not really neccesary nor propable,
			// it's against integer division by zero
		}

		/*
		 *	inline int CGPU_AdaBoost::TClassifierData::n_Phase(
		 *		const TClassifierData &r_t_classifier) const
		 *		- returns phase shift of r_t_classifier stipple mask
		 *		  against the one of this classifier
		 *		- note classifiers *must* be compatible
		 */
		inline int n_Phase(const TClassifierData &r_t_classifier) const
		{
			_ASSERTE(b_Compatible(r_t_classifier));
			return ((r_t_classifier.p_classifier->n_position_x -
				p_classifier->n_position_x) / p_classifier->n_block_width) % 4;
		}

		/*
		 *	inline bool b_CompareMask(int n_this_pass, int n_that_pass,
		 *		const TClassifierData &r_t_classifier) const
		 *		- compares one quarter of stipple mask of this
		 *		  classifier with the one of r_t_classifier
		 *		- n_this_pass and n_that_pass are mask indices (0 to 3)
		 *		- returns true if mask quarters are equal, false if they're not
		 */
		inline bool b_CompareMask(int n_this_pass, int n_that_pass,
			const TClassifierData &r_t_classifier) const
		{
			return !memcmp(p_stipple_mask[n_this_pass],
				r_t_classifier.p_stipple_mask[n_that_pass], 32 * 4 * sizeof(unsigned char));
		}
	};

	struct vec2 {
		float x, y;

		vec2()
		{}

		vec2(float _x, float _y)
			:x(_x), y(_y)
		{}

		inline bool operator ==(const vec2 &r_v_other) const
		{
			return r_v_other.x == x && r_v_other.y == y;
		}
	};

	struct vec4 {
		float x, y, z, w;

		vec4()
		{}

		vec4(float _x, float _y, float _z, float _w)
			:x(_x), y(_y), z(_z), w(_w)
		{}

		inline bool operator ==(const vec4 &r_v_other) const
		{
			return r_v_other.x == x && r_v_other.y == y && r_v_other.z == z && r_v_other.w == w;
		}
	};

	/*
	 *	struct TClassifierGroupShader
	 *		- contains information neccesary for executing group of classifiers
	 */
	struct TClassifierGroupShader {
		std::vector<const CGLTexture_3D*> texture_list;
		std::vector<CRasterizerCalibrator::TTexCoordRect> texcoord_list;
		CGLTexture_2D *p_alpha_texture;
		vec2 v_alpha_pixel; // coords to alpha texture start with .5 and rise with 1 pixel step
		std::vector<vec4> block_to_slice_list; // v_block_to_slice
		std::vector<vec2> pixel_list; // v_pixel
		std::vector<const TClassifierData*> classifier_list; // copy of classifiers in question
		TShaderInfo *p_shader_info[4]; // shader source code (output)

		typedef std::vector<const TClassifierData*>::iterator TClassifierIterator;

		TClassifierGroupShader(CGLState *p_state,
			TClassifierIterator p_begin, TClassifierIterator p_end);
		~TClassifierGroupShader();
		static const char *p_s_BlocktoSlice_Name(int n_index);
		static const char *p_s_Pixel_Name(int n_index);
		static const char *p_s_Texture_Name(int n_index);
		static const char *p_s_TexCoords_Name(int n_index);

	protected:
		bool Create(CGLState *p_state);
	};

	/*
	 *	struct TClassifierGroupRuntime
	 *		- more compact version of TClassifierGroupShader
	 *		  used for real rendering (not just verification)
	 */
	struct TClassifierGroupRuntime {
		std::vector<const CGLTexture_3D*> texture_list; // source textures
		CGLTexture_2D *p_alpha_texture; // alpha texture
		CGLShaderObject *p_shader[4]; // shaders
		int n_display_list; // geometry display list
		unsigned char p_stipple_mask[4][32 * 4]; // stipple masks

		TClassifierGroupRuntime(); // only creates empty TClassifierGroupRuntime, call Create
		~TClassifierGroupRuntime();
		bool Create(CGLState *p_state, TClassifierGroupShader &r_cgs); // call once!
		void Render(CGLState *p_state) const;
	};

	std::vector<TClassifierData> m_classifier_data;

	int m_n_group_size_limit;

	std::vector<TClassifierGroupRuntime*> m_runtime_info;
	CGLRenderBuffer_FBO *m_p_frame_buffer;
	CGLTexture_2D *m_p_result_texture; // GL_RGB16F_ARB, but red channel only is used

public:
	/*
	 *	CGPU_AdaBoost::CGPU_AdaBoost(CGLState *p_state, int n_image_width, int n_image_height,
	 *		const TAdaBoostClassifier &r_t_classifier, CConvolutionManager &r_conv_plan)
	 *		- default constructor
	 *		- creates AdaBoost processing plan
	 *		- p_state is OpenGL state guard (needed for rasterizer calibration)
	 *		- n_image_width and n_image_height are input image width and height
	 *		- r_t_classifier contains weak classifiers as loaded from xml
	 *		- r_conv_plan is convolution plan (contains convolution images neccessary
	 *		  to evaluate for those classifiers)
	 *		- if any error occured, classifier list is cleared (number of classifiers returns 0)
	 */
	CGPU_AdaBoost(CGLState *p_state, int n_image_width, int n_image_height,
		const TAdaBoostClassifier &r_t_classifier, CConvolutionManager &r_conv_plan,
		int n_group_size_limit = 24);

	/*
	 *	CGPU_AdaBoost::~CGPU_AdaBoost()
	 *		- destructor
	 */
	~CGPU_AdaBoost();

	/*
	 *	static int CGPU_AdaBoost::n_Max_VertexInterpolants()
	 *		- returns number of vertex interpolants
	 *		- OpenGL must be initialized before calling this
	 */
	static int n_Max_VertexInterpolants();

	/*
	 *	static int CGPU_AdaBoost::n_Max_ConvolutionTextures()
	 *		- returns number of textures available for convolution textures
	 *		- OpenGL must be initialized before calling this
	 */
	static int n_Max_ConvolutionTextures();

	/*
	 *	bool CGPU_AdaBoost::b_Status() const
	 *		- returns AdaBoost plan status; if plan is valid returns true, otherwise returns false
	 *		- note this reflects constructor outcome
	 */
	bool b_Status() const;

	/*
	 *	inline const CGLTexture_2D *CGPU_AdaBoost::p_Result_Texture() const
	 *		- returns texture with classification results
	 *		- contents are undefined unless ProcessFrame() is called
	 */
	inline const CGLTexture_2D *p_Result_Texture() const
	{
		return m_p_result_texture;
	}

	/*
	 *	int CGPU_AdaBoost::n_Classifier_Num() const
	 *		- returns number of weak classifier
	 *		- note this returns 0 if constructor failed
	 */
	int n_Classifier_Num() const;

	/*
	 *	bool CGPU_AdaBoost::ProcessFrame(CGLState *p_state)
	 *		- processes a single frame (call convolution manager before calling this)
	 *		- p_state is OpenGL state guard
	 *		- returns true if everything is ok and no OpenGL errors occur; otherwise returns false
	 */
	bool ProcessFrame(CGLState *p_state);

	// todo - document this
	bool Validate_ClassifierInputs(CGLState *p_state, CGLTexture_2D *p_source_image,
		CConvolutionManager &r_conv_plan, int &r_n_max_error, float &r_f_std_deviation);
	bool Validate_SliceSelection(CGLState *p_state, CGLTexture_2D *p_source_image,
		CConvolutionManager &r_conv_plan, int &r_n_slice_mismatch_num, float &r_f_mismatch_rate);
	bool Validate_FeatureSampling(CGLState *p_state, CGLTexture_2D *p_source_image,
		CConvolutionManager &r_conv_plan, bool &r_b_errors);
	bool Validate_ShaderSynthesis(CGLState *p_state, CGLTexture_2D *p_source_image,
		CConvolutionManager &r_conv_plan, bool &r_b_errors);

protected:
	bool TestShaderRun(CGLState *p_state,
		const TClassifierGroupShader &r_sh, CGLShaderObject *p_shader[4],
		CGLTexture_2D *p_result_texture, CGLRenderBuffer_FBO *p_render_buffer,
		unsigned char *p_aux_buffer0, float *p_aux_buffer1, float *p_aux_buffer2);

	bool CreateClassifierGroups(std::vector<const TClassifierData*> &r_classifier_group_list,
		std::vector<int> &r_classifier_group_first_idx);
	static bool CompileLinkReport(CGLState *p_state, CGLShaderObject &r_shader);
	static int n_AreaSum(const unsigned char *p_image, int n_width,
		int n_x0, int n_y0, int n_x1, int n_y1);
	static void DeleteRTI(TClassifierGroupRuntime *p_cgr);
};

#endif //__LRD_CLASSIFIER_GPU_ENGINE_INCLUDED
