/*
								+-----------------------------------+
								|                                   |
								|  ***  Convolution scheduler  ***  |
								|                                   |
								|   Copyright   -tHE SWINe- 2007   |
								|                                   |
								|         ConvSchedule.cpp          |
								|                                   |
								+-----------------------------------+
*/

#include <vector>
#include "../../berLame_src/CallStack.h"
#include <stdlib.h>
#include <stdio.h>
#include <algorithm>
#include <functional>
#ifdef WIN32
#include <windows.h> // gl.h needs some typedefs
#endif
#include <GL/gl.h>
#include <GL/glext.h>
#include "../../berLame_src/OpenGL20.h"
#include "../../berLame_src/OpenGLState.h"
#include "../../berLame_src/VertexBufferObject.h"
#include "../../berLame_src/Shader2.h"
#include "../../berLame_src/ShaderSys.h"
#include "../../berLame_src/Texture.h"
#include "../../berLame_src/RenderBuffer2.h"
#include "Load_XML.h"
#include "ConvSchedule.h"
#include "RasCalib.h"

#if defined(_MSC_VER) && !defined(__MWERKS__) && !defined(for)
#define for if(0) {} else for
#endif

/*
 *								=== TConvolutionInfo ===
 */

/*
 *	TConvolutionInfo::TConvolutionInfo()
 *		- default constructor
 */
TConvolutionInfo::TConvolutionInfo()
{}

/*
 *	TConvolutionInfo::TConvolutionInfo(const TConvolutionInfo &r_t_info)
 *		- copy constructor
 */
TConvolutionInfo::TConvolutionInfo(const TConvolutionInfo &r_t_info)
	:n_block_width(r_t_info.n_block_width), n_block_height(r_t_info.n_block_height),
	n_x_shift(r_t_info.n_x_shift), n_y_shift(r_t_info.n_y_shift)
{}

/*
 *	TConvolutionInfo::TConvolutionInfo(const TWeakClassifier &r_t_weak_cl)
 *		- constructor creating convolution info satisfying weak classifier demands
 */
TConvolutionInfo::TConvolutionInfo(const TWeakClassifier &r_t_weak_cl)
	:n_block_width(r_t_weak_cl.n_block_width), n_block_height(r_t_weak_cl.n_block_height),
	n_x_shift(r_t_weak_cl.n_position_x % r_t_weak_cl.n_block_width),
	n_y_shift(r_t_weak_cl.n_position_y % r_t_weak_cl.n_block_height)
{}

/*
 *	TConvolutionInfo TConvolutionInfo::operator =(const TConvolutionInfo &r_t_info)
 *		- copy operator
 */
TConvolutionInfo TConvolutionInfo::operator =(const TConvolutionInfo &r_t_info)
{
	n_block_width = r_t_info.n_block_width;
	n_block_height = r_t_info.n_block_height;
	n_x_shift = r_t_info.n_x_shift;
	n_y_shift = r_t_info.n_y_shift;
	return *this;
}

/*
 *	bool TConvolutionInfo::operator ==(const TConvolutionInfo &r_t_info)
 *		- equality operator
 */
bool TConvolutionInfo::operator ==(const TConvolutionInfo &r_t_info)
{
	return n_block_width == r_t_info.n_block_width &&
		n_block_height == r_t_info.n_block_height &&
		n_x_shift == r_t_info.n_x_shift &&
		n_y_shift == r_t_info.n_y_shift;
}

/*
 *	bool TConvolutionInfo::operator <(const TConvolutionInfo &r_t_info)
 *		- less-then operator for sorting
 */
bool TConvolutionInfo::operator <(const TConvolutionInfo &r_t_info)
{
	return n_block_width < r_t_info.n_block_width || (n_block_width == r_t_info.n_block_width &&
		(n_block_height < r_t_info.n_block_height || (n_block_height == r_t_info.n_block_height &&
		(n_x_shift < r_t_info.n_x_shift || (n_x_shift == r_t_info.n_x_shift &&
		n_y_shift < r_t_info.n_y_shift)))));
}

/*
 *	int TConvolutionInfo::Dump()
 *		- prints convolution info dump to stdout
 *		- MSVC can't handle return void; in templates so it returns number of prited chars
 */
int TConvolutionInfo::Dump()
{
	return printf("\tconvolution(block-width: %d, block-height: %d, shift-x: %d, shift-y: %d)\n",
		n_block_width, n_block_height, n_x_shift, n_y_shift);
}

/*
 *								=== ~TConvolutionInfo ===
 */

/*
 *								=== TConvolutionList ===
 */

/*
 *	void TConvolutionList::Dump()
 *		- prints convolution info dumps to stdout
 */
void TConvolutionList::Dump()
{
	printf("=== Convolution list dump ===\n");
	std::for_each(convolution_list.begin(), convolution_list.end(),
		std::mem_fun_ref(&TConvolutionInfo::Dump));
}

/*
 *	bool TConvolutionList::Create(int n_classifier_num, const TWeakClassifier *p_classifier)
 *		- creates list of convolutions necessary to evaluate given classifiers
 *		- n_classifier_num is number of classifiers, p_classifier is list of classifiers
 *		- returns true on success, false on failure
 */
bool TConvolutionList::Create(int n_classifier_num, const TWeakClassifier *p_classifier)
{
	convolution_list.erase(convolution_list.begin(), convolution_list.end());
	// erase original contents if any

	for(int i = 0; i < n_classifier_num; ++ i) {
		TConvolutionInfo t_conv_info(p_classifier[i]);
		if(std::find(convolution_list.begin(), convolution_list.end(),
		   t_conv_info) == convolution_list.end()) {
			convolution_list.reserve(convolution_list.size() + 1);
			if(convolution_list.capacity() < convolution_list.size() + 1)
				return false;
			convolution_list.push_back(t_conv_info);
		}
	}
	// simply create list of all different convolutions and shifts

	std::sort(convolution_list.begin(), convolution_list.end());
	// sort them by size

	return true;
}

/*
 *								=== ~TConvolutionList ===
 */

/*
 *								=== TShaderParameters ===
 */

TShaderParameters::TShaderParameters(CGLState *p_state,
	const TConvolutionInfo &r_t_info, int n_src_w, int n_src_h)
	:n_source_w(n_src_w), n_source_h(n_src_h),
	n_offset_x(r_t_info.n_x_shift), n_offset_y(r_t_info.n_y_shift),
	n_samples_x(r_t_info.n_block_width), n_samples_y(r_t_info.n_block_height),
	n_mip_level(0), p_render_buffer(0), p_image(0)
{
	while(!(n_offset_x & 1) && !(n_offset_y & 1) &&
	   !(n_samples_x & 1) && !(n_samples_y & 1) &&
	   !(n_source_w & 1) && !(n_source_h & 1)) {
		++ n_mip_level;
		n_offset_x /= 2;
		n_offset_y /= 2;
		n_samples_x /= 2;
		n_samples_y /= 2;
		n_source_w /= 2;
		n_source_h /= 2;
	}

	n_result_w = n_source_w / n_samples_x;
	n_result_w -= n_result_w % 4; // grey to RGBA pack
	n_result_h = n_source_h / n_samples_y;

	f_pixel_w = 1.0f / n_source_w;
	f_pixel_h = 1.0f / n_source_h;
	// calculate pixel size

	n_result_w /= 4;
	// grey to RGBA pack

	f_min_s = (n_offset_x + .25f) * f_pixel_w;
	f_min_t = (n_offset_y + .25f) * f_pixel_h; // sub-kernel offset (+ .25f to pixel center)
	f_max_s = f_min_s + (n_result_w - 1) * n_samples_x * 4 * f_pixel_w;
	f_max_t = f_min_t + (n_result_h - 1) * n_samples_y * f_pixel_h; // texture size is 1, 1
	// calculate texture coordinates

	int n_source_size = (n_source_w > n_source_h)? n_source_w : n_source_h;
	CRasterizerCalibrator::TTexCoordRect t_texcoords(f_min_s, f_min_t, f_max_s, f_max_t);
	if(CRasterizerCalibrator::CalibrateRasterizer(p_state, n_result_w, n_result_h,
	   t_texcoords, t_texcoords, .25f / n_source_size)) {
		f_min_s = t_texcoords.f_min_s;
		f_min_t = t_texcoords.f_min_t;
		f_max_s = t_texcoords.f_max_s;
		f_max_t = t_texcoords.f_max_t;

		n_source_w = n_src_w;
		n_source_h = n_src_h;
		// remember original source dimensions
	} else {
		n_source_w = -1;
		n_source_h = -1;
		// set invalid source dimensions to mark error
	}
	// calibrate rasterizer

	n_shader_type = (n_samples_x == 1 && n_samples_y == 1)? conv_PackShader : conv_ConvShader;
	// determine shader type
}

TShaderParameters::TShaderParameters(const TShaderParameters &r_t_sp)
	:n_shader_type(r_t_sp.n_shader_type), n_offset_x(r_t_sp.n_offset_x),
	n_offset_y(r_t_sp.n_offset_y), n_mip_level(r_t_sp.n_mip_level),
	n_samples_x(r_t_sp.n_samples_x), n_samples_y(r_t_sp.n_samples_y),
	f_pixel_w(r_t_sp.f_pixel_w), f_pixel_h(r_t_sp.f_pixel_h), f_min_s(r_t_sp.f_min_s),
	f_min_t(r_t_sp.f_min_t), f_max_s(r_t_sp.f_max_s), f_max_t(r_t_sp.f_max_t),
	n_source_w(r_t_sp.n_source_w), n_source_h(r_t_sp.n_source_h),
	n_result_w(r_t_sp.n_result_w), n_result_h(r_t_sp.n_result_h),
	p_render_buffer(r_t_sp.p_render_buffer), p_image(r_t_sp.p_image)
{}

TShaderParameters TShaderParameters::operator =(const TShaderParameters &r_t_sp)
{
	n_shader_type = r_t_sp.n_shader_type;
	n_offset_x = r_t_sp.n_offset_x;
	n_offset_y = r_t_sp.n_offset_y;
	n_mip_level = r_t_sp.n_mip_level;
	n_samples_x = r_t_sp.n_samples_x;
	n_samples_y = r_t_sp.n_samples_y;
	f_pixel_w = r_t_sp.f_pixel_w;
	f_pixel_h = r_t_sp.f_pixel_h;
	f_min_s = r_t_sp.f_min_s;
	f_min_t = r_t_sp.f_min_t;
	f_max_s = r_t_sp.f_max_s;
	f_max_t = r_t_sp.f_max_t;
	n_source_w = r_t_sp.n_source_w;
	n_source_h = r_t_sp.n_source_h;
	n_result_w = r_t_sp.n_result_w;
	n_result_h = r_t_sp.n_result_h;
	p_render_buffer = r_t_sp.p_render_buffer;
	p_image = r_t_sp.p_image;

	return *this;
}

void TShaderParameters::Dump()
{
	printf("\tshader_params(type: \'%s\', offset-x: %d offset-y: %d,"
		" mip-level: %d, samples-x:%d, samples-y: %d)\n",
		(n_shader_type == conv_PackShader)? "pack" : "conv", n_offset_x, n_offset_y,
		n_mip_level, n_samples_x, n_samples_y);
}

/*
 *								=== ~TShaderParameters ===
 */

/*
 *								=== CAdHocShaderInfo ===
 */

CAdHocShaderInfo::CAdHocShaderInfo(const char *p_s_shader_name, bool b_fragment_shader,
	const char *p_s_source_code, const char *p_s_sampler_names)
{
	p_s_name = (char*)p_s_shader_name;
	b_fragment = b_fragment_shader; // vertex or fragment?
	b_shader_target = true; // "shader", not "program"
	p_low = 0; // no program
	p_high = &m_t_high; // my shader

	m_t_high.p_s_source_code = (char*)p_s_source_code;
	m_t_high.n_sampler_num = 0;
	if(p_s_sampler_names) {
		for(const char *p_s_sampler = p_s_sampler_names; p_s_sampler;
		   p_s_sampler = (strchr(p_s_sampler, '|'))? strchr(p_s_sampler, '|') + 1 : 0)
			++ m_t_high.n_sampler_num;
		// count samplers

		m_t_high.p_sampler = new TShaderInfo::TShader::TSamplerInfo[m_t_high.n_sampler_num];
		_ASSERTE(m_t_high.p_sampler);
		// alloc samplers

		m_t_high.n_sampler_num = 0;
		for(const char *p_s_sampler = p_s_sampler_names; p_s_sampler;
		   p_s_sampler = (strchr(p_s_sampler, '|'))? strchr(p_s_sampler, '|') + 1 : 0) {
			char *p_s_str;
			if(strchr(p_s_sampler, '|')) {
				int n_size = strchr(p_s_sampler, '|') - p_s_sampler;
				if(p_s_str = new char[n_size + 1]) {
					strncpy(p_s_str, p_s_sampler, n_size);
					p_s_str[n_size] = 0;
				}
			} else
				p_s_str = p_s_Alloc(p_s_sampler);
			_ASSERTE(p_s_str);
			// get a single sampler name

			m_t_high.p_sampler[m_t_high.n_sampler_num].p_s_name = p_s_str;
			m_t_high.p_sampler[m_t_high.n_sampler_num].n_resource_index = m_t_high.n_sampler_num;
			// assign samplers

			++ m_t_high.n_sampler_num;
		}
		// assign sampler names and indices
	} else
		m_t_high.p_sampler = 0;
}

CAdHocShaderInfo::~CAdHocShaderInfo()
{
	m_t_high.p_s_source_code = 0;
	// don't want to be deleted by TShaderInfo::TShader destructor as it's static data

	p_s_name = 0;
	p_high = 0;
	// don't want them to be deleted by TShaderInfo destructor as they are static data
}

/*
 *								=== ~CAdHocShaderInfo ===
 */

/*
 *								=== CConvolutionManager ===
 */

CConvolutionManager::CConvolutionManager()
	:m_p_pack_shader(0), m_p_convolve_shader(0)
{}

CConvolutionManager::~CConvolutionManager()
{
	if(m_p_pack_shader)
		delete m_p_pack_shader;
	if(m_p_convolve_shader)
		delete m_p_convolve_shader;
	std::for_each(m_render_buffer_list.begin(), m_render_buffer_list.end(), Delete_FBO);
	std::for_each(m_image_list.begin(), m_image_list.end(), Delete_Texture);
}

bool CConvolutionManager::CreateConvolutionPlan(CGLState *p_state,
	const TConvolutionList &r_t_conv_list, int n_source_width, int n_source_height,
	bool b_16_bit_sums)
{
	m_shader_params_list.erase(m_shader_params_list.begin(), m_shader_params_list.end());
	// delete shader parameters

	std::for_each(m_render_buffer_list.begin(), m_render_buffer_list.end(), Delete_FBO);
	m_render_buffer_list.erase(m_render_buffer_list.begin(), m_render_buffer_list.end());
	std::for_each(m_image_list.begin(), m_image_list.end(), Delete_Texture);
	m_image_list.erase(m_image_list.begin(), m_image_list.end());
	// delete FBO's and textures

	m_shader_params_list.reserve(r_t_conv_list.convolution_list.size());
	if(m_shader_params_list.capacity() < r_t_conv_list.convolution_list.size())
		return false;
	m_image_list.reserve(r_t_conv_list.convolution_list.size());
	if(m_image_list.capacity() < r_t_conv_list.convolution_list.size())
		return false;

	for(int i = 0; i < r_t_conv_list.convolution_list.size(); ++ i) {
		TShaderParameters t_sh_p(p_state, r_t_conv_list.convolution_list[i],
			n_source_width, n_source_height);
		// create shader parameters

		if(t_sh_p.n_source_w < 0) {
			_ASSERTE(n_source_width > 0);
			return false;
		}
		// failed to calibrate

		std::vector<CGLRenderBuffer_FBO*>::const_iterator p_rb_iter;
		if((p_rb_iter = std::find_if(m_render_buffer_list.begin(), m_render_buffer_list.end(),
		   CFindGLResource(t_sh_p.n_result_w, t_sh_p.n_result_h, (b_16_bit_sums)?
		   GL_RGBA16 : GL_RGBA8))) == m_render_buffer_list.end()) {
			m_render_buffer_list.reserve(m_render_buffer_list.size() + 1);
			if(m_render_buffer_list.capacity() < m_render_buffer_list.size() + 1)
				return false;
			// make sure there is enough space in the list

			CGLRenderBuffer_FBO *p_render_buffer;
			if(!(p_render_buffer = new CGLRenderBuffer_FBO(t_sh_p.n_result_w,
			   t_sh_p.n_result_h, 1, true, (b_16_bit_sums)? GL_RGBA16 : GL_RGBA8,
			   false, false, 0, false, false, 0)))
				return false;
			if(!p_render_buffer->b_Status()) {
				delete p_render_buffer;
				return false;
			}
			// create a new render buffer

			t_sh_p.p_render_buffer = p_render_buffer;
			m_render_buffer_list.push_back(p_render_buffer);
		} else
			t_sh_p.p_render_buffer = *p_rb_iter;
		// try to find suitable framebuffer or to create one

		CGLTexture_2D *p_texture;
		if(!(p_texture = new CGLTexture_2D(p_state, t_sh_p.n_result_w, t_sh_p.n_result_h,
		   (b_16_bit_sums)? GL_RGBA16 : GL_RGBA8, true, 0)))
			return false;
		p_texture->r_Parameters(p_state).Set_Texture_Wrap_S(GL_CLAMP_TO_BORDER);
		p_texture->r_Parameters(p_state).Set_Texture_Wrap_T(GL_CLAMP_TO_BORDER);
		p_texture->r_Parameters(p_state).Set_Texture_Mag_Filter(GL_NEAREST);
		p_texture->r_Parameters(p_state).Set_Texture_Min_Filter(GL_NEAREST_MIPMAP_NEAREST);
		if(!p_texture->b_Status() || glGetError() != GL_NO_ERROR) {
			delete p_texture;
			return false;
		}
		m_image_list.push_back(t_sh_p.p_image = p_texture);
		// create texture to hold convolution results

		m_shader_params_list.push_back(t_sh_p);
	}
	// create list with shader parameters

	if(!m_p_pack_shader) {
		CAdHocShaderInfo pack_shader("pack_shader", true,
			"uniform sampler2D n_texture;\n"
			"//uniform vec2 v_pixel_align;\n" // 4 * pixel-w, pixel-h
			"uniform vec4 v_block_shift;\n" // pixel-w, 2 * pixel-w, 3 * pixel-w, 0
			"\n"
			"void main()\n"
			"{\n"
			"    vec4 v_sample;\n"
			"    vec2 v_texcoord = gl_TexCoord[0].xy;// - mod(gl_TexCoord[0].xy, v_pixel_align);\n"
			"    v_sample.x = texture2D(n_texture, v_texcoord).x;\n"
			"    v_sample.y = texture2D(n_texture, v_texcoord + v_block_shift.xw).x;\n"
			"    v_sample.z = texture2D(n_texture, v_texcoord + v_block_shift.yw).x;\n"
			"    v_sample.w = texture2D(n_texture, v_texcoord + v_block_shift.zw).x;\n" // pack four grey pixels to RGBA pixel
			"    gl_FragColor = v_sample;\n"
			"}\n");
		if(!(m_p_pack_shader = new CGLShaderObject(0, &pack_shader)))
			return false;
		char *p_s_compile = 0, *p_s_link = 0;
		bool b_result = m_p_pack_shader->Compile_Link(p_state, p_s_compile, p_s_link);
		if(p_s_compile) {
			printf("\n=== Pack shader compiler report ===\n%s\n", p_s_compile);
			delete[] p_s_compile;
		}
		if(p_s_link) {
			printf("\n=== Pack shader linker report ===\n%s\n", p_s_link);
			delete[] p_s_link;
		}
		if(!b_result)
			return false;
	}

	if(!m_p_convolve_shader) {
		CAdHocShaderInfo convolve_shader("convolve_shader", true,
			"uniform sampler2D n_texture;\n"
			"uniform vec2 v_pixel;\n" // pixel-w, pixel-h
			"//uniform vec2 v_pixel_align;\n" // 4 * pixel-w, pixel-h
			"uniform ivec2 v_kernel;\n" // width, height
			"uniform vec4 v_block_shift;\n" // width * pixel-w, 2 * width * pixel-w, 3 * width * pixel-w, 0
			"uniform float f_rescale;\n" // 1.0f / (width * height)
			"\n"
			"void main()\n"
			"{\n"
			"    vec4 v_sum = vec4(.0, .0, .0, .0);\n"
			"    vec2 v_texcoord = gl_TexCoord[0].xy;// - mod(gl_TexCoord[0].xy, v_pixel_align);\n"
			"    float f_restore_x = v_texcoord.x;\n"
			"    for(int i = 0; i != v_kernel.x; ++ i) {\n"
			"        for(int j = 0; j != v_kernel.y; ++ j) {\n"
			"            vec2 v_tc = v_texcoord + v_pixel * vec2(i, j);\n"
			"            vec4 v_sample;\n"
			"            v_sample.x = texture2D(n_texture, v_tc).x;\n"
			"            v_sample.y = texture2D(n_texture, v_tc + v_block_shift.xw).x;\n"
			"            v_sample.z = texture2D(n_texture, v_tc + v_block_shift.yw).x;\n"
			"            v_sample.w = texture2D(n_texture, v_tc + v_block_shift.zw).x;\n" // calculate sum of four blocks at once
			"            v_sum += v_sample\n;"
			"        }\n"
			"    }\n"
			"    gl_FragColor = v_sum * f_rescale;\n"
			"}\n");
		if(!(m_p_convolve_shader = new CGLShaderObject(0, &convolve_shader)))
			return false;
		char *p_s_compile = 0, *p_s_link = 0;
		bool b_result = m_p_convolve_shader->Compile_Link(p_state, p_s_compile, p_s_link);
		if(p_s_compile) {
			printf("\n=== Convolve shader compiler report ===\n%s\n", p_s_compile);
			delete[] p_s_compile;
		}
		if(p_s_link) {
			printf("\n=== Convolve shader linker report ===\n%s\n", p_s_link);
			delete[] p_s_link;
		}
		if(!b_result)
			return false;
	}

	return true;
}

bool CConvolutionManager::Convolve(CGLState *p_state, CGLTexture_2D *p_source_image)
{
	if(!m_p_convolve_shader || !m_p_convolve_shader->b_Status() ||
	   !m_p_pack_shader || !m_p_pack_shader->b_Status())
		return false;
	// shaders aren't ready

	glMatrixMode(GL_TEXTURE);
	glLoadIdentity();
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	// reset matrices

	p_source_image->r_Parameters(p_state).Set_Texture_Min_Filter(GL_NEAREST);
	p_source_image->r_Parameters(p_state).Set_Texture_Mag_Filter(GL_NEAREST);
	/*p_source_image->r_Parameters(p_state).Set_Texture_Min_Filter(GL_NEAREST_MIPMAP_LINEAR);
	p_source_image->r_Parameters(p_state).Set_Texture_Mag_Filter(GL_LINEAR);*/
	p_source_image->r_Parameters(p_state).Set_Texture_Wrap_S(GL_CLAMP_TO_BORDER);
	p_source_image->r_Parameters(p_state).Set_Texture_Wrap_T(GL_CLAMP_TO_BORDER);
	// make sure the texture has correct filtering

	int p_orig_viewport[4];
	glGetIntegerv(GL_VIEWPORT, p_orig_viewport);

	for(int i = 0; i < m_shader_params_list.size(); ++ i) {
		TShaderParameters t_params = m_shader_params_list[i];

		if(t_params.n_source_w != p_source_image->n_Width() ||
		   t_params.n_source_h != p_source_image->n_Height())
			return false;

		_ASSERTE(std::find(m_render_buffer_list.begin(), m_render_buffer_list.end(),
			t_params.p_render_buffer) != m_render_buffer_list.end());
		_ASSERTE(std::find(m_image_list.begin(), m_image_list.end(),
			t_params.p_image) != m_image_list.end());

		t_params.p_render_buffer->Bind();
		t_params.p_render_buffer->BindColorTexture_2D(*t_params.p_image, GL_TEXTURE_2D);
		glViewport(0, 0, t_params.p_render_buffer->n_Width(), t_params.p_render_buffer->n_Height());
		// bind texture for rendering

		glClear(GL_COLOR_BUFFER_BIT);

		//p_state = t_params.p_render_buffer->p_State(); // we're using FBO's, they don't have it's own state
		p_source_image->Bind_Enable(p_state);
		p_source_image->r_Parameters(p_state).Set_Texture_Base_Level(t_params.n_mip_level);
		p_source_image->r_Parameters(p_state).Set_Texture_Max_Level(t_params.n_mip_level);
		p_state->DisableDepthTest();
		p_state->DisableCullFace();
		p_state->DisableBlend();
		p_state->DisableAlphaTest();
		//p_state->Color3f(1, 1, 1); // doesn't matter, shaders ignore it
		// setup GL state

		if(t_params.n_shader_type == TShaderParameters::conv_PackShader) {
			_ASSERTE(t_params.n_samples_x == 1 && t_params.n_samples_y == 1);
			m_p_pack_shader->t_Find_Parameter("v_block_shift").SetValue(t_params.f_pixel_w,
				2 * t_params.f_pixel_w, 3 * t_params.f_pixel_w, 0);
			//m_p_pack_shader->t_Find_Parameter("v_pixel_align").SetValue(4 * t_params.f_pixel_w, t_params.f_pixel_h);
			m_p_pack_shader->Bind(p_state);
		} else {
			//m_p_convolve_shader->t_Find_Parameter("v_pixel_align").SetValue(4 * t_params.f_pixel_w, t_params.f_pixel_h);
			m_p_convolve_shader->t_Find_Parameter("v_block_shift").SetValue(t_params.f_pixel_w * t_params.n_samples_x,
				2 * t_params.n_samples_x * t_params.f_pixel_w, 3 * t_params.n_samples_x * t_params.f_pixel_w, 0);
			m_p_convolve_shader->t_Find_Parameter("v_pixel").SetValue(t_params.f_pixel_w, t_params.f_pixel_h);
			m_p_convolve_shader->t_Find_Parameter("f_rescale").SetValue(1.0f / (t_params.n_samples_x * t_params.n_samples_y));
			m_p_convolve_shader->t_Find_IntParameter("v_kernel").SetValue(t_params.n_samples_x, t_params.n_samples_y);
			m_p_convolve_shader->Bind(p_state);
		}
		// bind shader

		glBegin(GL_QUADS);
		glTexCoord2f(t_params.f_min_s, t_params.f_max_t);
		glVertex2f(-1, 1);
		glTexCoord2f(t_params.f_max_s, t_params.f_max_t);
		glVertex2f(1, 1);
		glTexCoord2f(t_params.f_max_s, t_params.f_min_t);
		glVertex2f(1, -1);
		glTexCoord2f(t_params.f_min_s, t_params.f_min_t);
		glVertex2f(-1, -1);
		glEnd();
		// draw fullscreen quad to execute shader

		t_params.p_render_buffer->BindColorTexture_2D(0, GL_TEXTURE_2D);
		t_params.p_render_buffer->Release();
		// release texture and render buffer
	}

	p_source_image->r_Parameters(p_state).Set_Texture_Base_Level(0);
	p_source_image->r_Parameters(p_state).Set_Texture_Max_Level(1000);
	// put back normal mip level

	if(m_p_pack_shader->n_Api() == shader_api_Core_shader)
		p_state->BindProgramObject_GL2(0);
	else
		p_state->BindProgramObject(0);
	// bind no shader

	glViewport(p_orig_viewport[0], p_orig_viewport[1], p_orig_viewport[2], p_orig_viewport[3]);
	// restore original viewport

	return glGetError() == GL_NO_ERROR;
}

void CConvolutionManager::Delete_FBO(CGLRenderBuffer_FBO *p_fbo)
{
	delete p_fbo;
}

void CConvolutionManager::Delete_Texture(CGLTexture_2D *p_tex)
{
	delete p_tex;
}

/*
 *								=== ~CConvolutionManager ===
 */
